# -*- coding: utf-8 -*-
#
# metadata.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     27 January 2014
# Copyright (c) 2014-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import re
import subprocess

from flent import util
from flent.loggers import get_logger

logger = get_logger(__name__)

TIMEOUT = 5
INTERESTING_OFFLOADS = ['tcp-segmentation-offload',
                        'udp-fragmentation-offload',
                        'large-receive-offload',
                        'generic-segmentation-offload',
                        'generic-receive-offload']

INTERESTING_SYSCTLS = ['net.core.rmem_max',
                       'net.core.wmem_max',
                       'net.ipv4.tcp_autocorking',
                       'net.ipv4.tcp_early_retrans',
                       'net.ipv4.tcp_ecn',
                       'net.ipv4.tcp_pacing_ca_ratio',
                       'net.ipv4.tcp_pacing_ss_ratio',
                       'net.ipv4.tcp_dsack',
                       'net.ipv4.tcp_fack',
                       'net.ipv4.tcp_sack',
                       'net.ipv4.tcp_fastopen',
                       'net.ipv4.tcp_syncookies',
                       'net.ipv4.tcp_window_scaling',
                       'net.ipv4.tcp_notsent_lowat',
                       'net.ipv4.tcp_limit_output_bytes',
                       'net.ipv4.tcp_timestamps',
                       'net.ipv4.tcp_congestion_control',
                       'net.ipv4.tcp_allowed_congestion_control',
                       'net.ipv4.tcp_available_congestion_control',
                       'net.ipv4.tcp_mem',
                       'net.ipv4.tcp_rmem',
                       'net.ipv4.tcp_wmem',
                       'net.ipv4.tcp_moderate_rcvbuf',
                       'net.ipv4.tcp_no_metrics_save']

# Modules we try to get versions for
INTERESTING_MODULES = ['cake',
                       'ath',
                       'ath9k',
                       'ath9k_common',
                       'ath9k_hw',
                       'mac80211',
                       'cfg80211']

try:
    processerror = subprocess.SubprocessError
except AttributeError:
    processerror = subprocess.CalledProcessError


class CommandRunner(object):

    def __init__(self):
        self.hostname = None
        self.env = os.environ.copy()
        self.fixup_path()

    def fixup_path(self):
        """Fix up the PATH to include /sbin and /usr/sbin since some of the
        utilities called (such as ip and tc) live here, and those directories
        are not normally in the path (on e.g. Debian)."""
        path = self.env['PATH'].split(':')
        if '/sbin' not in path:
            path.append('/sbin')
        if '/usr/sbin' not in path:
            path.append('/usr/sbin')
        self.env['PATH'] = ":".join(path)

    def set_hostname(self, hostname):
        self.hostname = hostname

    def __call__(self, command):
        """Try executing a command, and if successful,
        return the strip()'ed output, else None."""
        try:
            if self.hostname:
                logger.debug("Executing '%s' on host '%s'", command, self.hostname)
                command = "ssh %s '%s'" % (self.hostname, command)
            else:
                logger.debug("Executing '%s' on localhost", command)

            try:
                res = subprocess.check_output(command, universal_newlines=True,
                                              shell=True, stderr=subprocess.STDOUT,
                                              env=self.env, timeout=TIMEOUT)
            except TypeError:
                # Python 2 doesn't have timeout arg
                res = subprocess.check_output(command, universal_newlines=True,
                                              shell=True, stderr=subprocess.STDOUT,
                                              env=self.env)
            return res.strip()
        except processerror:
            return None


get_command_output = CommandRunner()

__all__ = ['record_metadata']


def record_metadata(results, extended, hostnames):
    logger.debug("Gathering local metadata")
    m = results.meta()
    get_command_output.set_hostname(None)
    m['KERNEL_NAME'] = get_command_output("uname -s")
    m['KERNEL_RELEASE'] = get_command_output("uname -r")
    m['MODULE_VERSIONS'] = get_module_versions()
    m['SYSCTLS'] = get_sysctls()
    m['EGRESS_INFO'] = get_egress_info(target=m['HOST'],
                                       ip_version=m['IP_VERSION'],
                                       extended=extended)

    if extended:
        m['IP_ADDRS'] = get_ip_addrs()
        m['GATEWAYS'] = get_gateways()
        m['WIFI_DATA'] = get_wifi_data()

    m['REMOTE_METADATA'] = {}

    for h in hostnames:
        logger.debug("Gathering remote metadata from %s", h)
        get_command_output.set_hostname(h)
        m['REMOTE_METADATA'][h] = {}
        m['REMOTE_METADATA'][h]['LOCAL_HOST'] = get_command_output("hostname")
        m['REMOTE_METADATA'][h]['KERNEL_NAME'] = get_command_output("uname -s")
        m['REMOTE_METADATA'][h]['KERNEL_RELEASE'] = get_command_output("uname -r")
        m['REMOTE_METADATA'][h]['MODULE_VERSIONS'] = get_module_versions()
        m['REMOTE_METADATA'][h]['SYSCTLS'] = get_sysctls()
        m['REMOTE_METADATA'][h]['EGRESS_INFO'] = get_egress_info(
            target=m['HOST'], ip_version=m['IP_VERSION'], extended=extended)

        if m['EGRESS_INFO'] is not None and 'src' in m['EGRESS_INFO']:
            m['REMOTE_METADATA'][h]['INGRESS_INFO'] = get_egress_info(
                target=m['EGRESS_INFO']['src'], ip_version=m['IP_VERSION'],
                extended=extended)
        else:
            m['REMOTE_METADATA'][h]['INGRESS_INFO'] = None
            m['REMOTE_METADATA'][h]['EGRESS_INFO'] = get_egress_info(
                target=m['HOST'], ip_version=m['IP_VERSION'],
                extended=extended)

        if extended:
            m['REMOTE_METADATA'][h]['IP_ADDRS'] = get_ip_addrs()
            m['REMOTE_METADATA'][h]['GATEWAYS'] = get_gateways()
            m['REMOTE_METADATA'][h]['WIFI_DATA'] = get_wifi_data()


def record_postrun_metadata(results, extended, hostnames):
    logger.debug("Recording postrun metadata")
    m = results.meta()
    get_command_output.set_hostname(None)
    if m['EGRESS_INFO'] is not None:

        m['EGRESS_INFO']['tc_stats_post'] = get_tc_stats(
            m['EGRESS_INFO']['iface'])

    for h in hostnames:
        get_command_output.set_hostname(h)
        for i in 'EGRESS_INFO', 'INGRESS_INFO':
            if m['REMOTE_METADATA'][h][i] is not None:
                m['REMOTE_METADATA'][h][i]['tc_stats_post'] = get_tc_stats(
                    m['REMOTE_METADATA'][h][i]['iface'])


def get_ip_addrs(iface=None):
    """Try to get IP addresses associated to this machine. Uses iproute2 if available,
    otherwise falls back to ifconfig."""
    addresses = {}

    cmd = "ip addr show"
    if iface is not None:
        cmd += " dev %s" % iface
    output = get_command_output(cmd)

    if output is None:
        cmd = "ifconfig"
        if iface is not None:
            cmd += " %s" % iface
        output = get_command_output(cmd)

    iface_re = re.compile(r'^([0-9]+: )?([a-z0-9-]+):')

    if output is not None:
        lines = output.splitlines()
        iface = None
        addrs = []
        for l in lines:
            # Both ifconfig and iproute2 emit addresses on lines starting with
            # the address identifier, and fields are whitespace-separated. Look
            # for that and parse accordingly.
            m = iface_re.match(l)
            if m is not None:
                if iface and addrs:
                    addresses[iface] = addrs
                iface = m.group(2)
                addrs = []
            parts = l.strip().split()
            if parts and parts[0] in ('inet', 'inet6'):
                a = parts[1]
                if '/' in a:  # iproute2 adds subnet qualification; strip that
                    a = a[:a.index('/')]
                if '%' in a:  # BSD may add interface qualification; strip that
                    a = a[:a.index('%')]
                addrs.append(a)
        if addrs and iface:
            addresses[iface] = addrs
    return addresses or None


def get_link_params(iface):
    link_params = {}
    output = get_command_output("ip link show dev %s" % iface)

    if output is None:
        output = get_command_output("ifconfig %s" % iface)

    if output is not None:
        m = re.search(r"(qlen|txqueuelen) (\d+)", output)
        if m:
            link_params['qlen'] = m.group(2)
        m = re.search(r"ether ([0-9a-f:]{17})", output)
        if m:
            link_params['ether'] = m.group(1)

    output = get_command_output("ethtool %s" % iface)
    if output is not None:
        m = re.search(r"Speed: ([0-9]+Mb/s)", output)
        if m:
            link_params['speed'] = m.group(1)
        m = re.search(r"Duplex: (\w+)", output)
        if m:
            link_params['duplex'] = m.group(1)

    return link_params or None


def get_offloads(iface):
    offloads = {}

    output = get_command_output("ethtool -k %s" % iface)
    val_map = {'on': True, 'off': False}
    if output is not None:
        for l in output.splitlines():
            parts = l.split()
            key = parts[0].strip(":")
            if key in INTERESTING_OFFLOADS:
                try:
                    offloads[key] = val_map[parts[1]]
                except KeyError:
                    continue
    return offloads or None


def get_gateways():
    gws = []
    # Linux netstat only outputs IPv4 data by default, but can be made to output
    # both if passed both -4 and -6
    output = get_command_output("netstat -46nr")
    if output is None:
        # If that didn't work, maybe netstat doesn't support -4/-6 (e.g. BSD),
        # so try without
        output = get_command_output("netstat -nr")
    if output is not None:
        output = output.replace("Next Hop", "Next_Hop")  # breaks part splitting
        lines = output.splitlines()
        iface_idx = None

        for line in lines:
            parts = line.split()
            if not parts:
                continue

            # Try to find the column header; should have "Destination" as first
            # word.
            if parts[0] == "Destination":
                # Linux uses Iface or If as header (for IPv4/6), FreeBSD uses If
                for n in ("Iface", "Netif", "If"):
                    if n in parts:
                        iface_idx = parts.index(n)
            if parts[0] in ("0.0.0.0", "default", "::/0"):
                if iface_idx is not None:
                    # The fields may run into each other in some instances; try
                    # to detect this, and if so just assume that the interface
                    # name is the last field (it often is, on Linux).
                    if iface_idx > len(parts) - 1:
                        iface_idx = -1
                    gw = {'ip': parts[1], 'iface': parts[iface_idx]}
                    if not gw['iface'].startswith('lo'):
                        gws.append(gw)
                else:
                    gws.append({'ip': parts[1]})
    return gws


def get_egress_info(target, ip_version, extended):
    route = {}

    if target:
        ip = util.lookup_host(target, ip_version)[4][0]
        output = get_command_output("ip route get %s" % ip)
        if output is not None:
            # Linux iproute2 syntax. Example:
            # $ ip r get 8.8.8.8
            # 8.8.8.8 via 10.109.3.254 dev wlan0  src 10.109.0.146
            #     cache
            parts = iter(output.split())
            for p in parts:
                if p == 'via':
                    route['nexthop'] = next(parts)
                elif p == 'dev':
                    route['iface'] = next(parts)
                elif p == 'src':
                    route['src'] = next(parts)
        else:
            output = get_command_output("route -n get %s" % ip)
            if output is not None:
                # BSD syntax. Example:
                # $ route -n get 8.8.8.8
                #    route to: 8.8.8.8
                # destination: default
                #        mask: default
                #     gateway: 10.42.7.225
                #   interface: em0
                #       flags: <UP,GATEWAY,DONE,STATIC>
                #  recvpipe  sendpipe  ssthresh  rtt,msec    mtu        weight    expire
                #        0         0         0         0      1500         1         0

                for line in output.splitlines():
                    if ":" not in line:
                        continue
                    k, v = [i.strip() for i in line.split(":")]
                    if k == "gateway":
                        route['nexthop'] = v
                    if k == "interface":
                        route['iface'] = v

    if route:
        route['qdiscs'] = get_qdiscs(route['iface'])
        route['tc_stats_pre'] = get_tc_stats(route['iface'])
        route['classes'] = get_classes(route['iface'])
        route['offloads'] = get_offloads(route['iface'])
        route['bql'] = get_bql(route['iface'])
        route['driver'] = get_driver(route['iface'])
        route['link_params'] = get_link_params(route['iface'])
        route['target'] = ip
        if 'nexthop' not in route:
            route['nexthop'] = None

        if not extended:
            for k in 'gateway', 'src', 'nexthop', 'target':
                if k in route:
                    del route[k]
            if route['link_params'] and 'ether' in route['link_params']:
                del route['link_params']['ether']

    return route or None


def parse_tc(cmd, kind):
    items = []

    output = get_command_output(cmd)
    if output is not None:
        lines = output.splitlines()
        for line in lines:
            if line.startswith(" "):
                itm = items[-1]
                if 'stats' in itm:
                    itm['stats'].append(line.strip())
                else:
                    itm['stats'] = [line.strip()]
                continue

            parts = line.split()
            if not parts or parts[0] != kind:
                continue
            item = {'name': parts[1],
                    'id': parts[2]}
            if parts[3] == 'root':
                item['parent'] = 'root'
                params = parts[4:]
            else:
                item['parent'] = parts[4]
                params = parts[5:]

            # Assume that the remainder of the output line is a set of space
            # delimited key/value pairs. Some qdiscs (e.g. fq_codel) has a
            # single non-valued parameter at the end, in which case the length
            # of params will be uneven. In this case an empty string is added as
            # the parameter "value", to make sure it is included.
            if len(params) % 2 > 0:
                params.append("")
            item['params'] = dict(zip(params[::2], params[1::2]))

            items.append(item)
    return items or None


def get_qdiscs(iface):
    return parse_tc("tc qdisc show dev %s" % iface, "qdisc")


def get_tc_stats(iface):
    output = get_command_output("tc -s qdisc show dev %s" % iface)
    items = []
    if output is not None:
        item = []
        # Split out output so we get one list entry for each qdisc -- first line
        # of a qdisc's stats output is non-indented, subsequent lines are
        # indented by spaces.
        for line in filter(None, output.splitlines()):
            if line.startswith(" "):
                item.append(line)
            else:
                if item:
                    items.append("\n".join(item))
                item = [line]
        if item:
            items.append("\n".join(item))
    return items or None


def get_classes(iface):
    return parse_tc("tc class show dev %s" % iface, "class")


def get_bql(iface):
    bql = []
    output = get_command_output(
        'for i in /sys/class/net/%s/queues/tx-*; do [ -d $i/byte_queue_limits ] '
        '&& echo -n "$(basename $i) " && cat $i/byte_queue_limits/limit_max; done'
        % iface)
    if output is not None:
        bql = dict([i.split() for i in output.splitlines()])

    return bql or None


def get_driver(iface):
    return get_command_output(
        "basename $(readlink /sys/class/net/%s/device/driver)" % iface)


def get_sysctls():
    sysctls = {}

    output = get_command_output("sysctl -e %s" % " ".join(INTERESTING_SYSCTLS))
    if output is not None:
        for line in output.splitlines():
            parts = line.split("=")
            if len(parts) != 2:
                continue
            k, v = [i.strip() for i in parts]
            try:
                sysctls[k] = int(v)
            except ValueError:
                sysctls[k] = v

    return sysctls


def get_module_versions():
    module_versions = {}
    modules = []

    output = get_command_output("find /sys/module -name .note.gnu.build-id")

    if output is not None:
        module_files = output.split()

        for f in module_files:
            if "/sections/" in f:
                continue
            m = f.replace("/sys/module/", "").split("/", 1)[0]
            if m in INTERESTING_MODULES:
                modules.append((m, f))

    if modules:

        # The hexdump output will be a string of hexadecimal values of the
        # concatenation of all the .note.gnu.build-id files.
        #
        # Each file starts with "040000001400000003000000474e5500" (0x474e550 is
        # "GNU\0"), so simply split on that to get the data we are interested in.
        version_strings = get_command_output(
            "hexdump -ve \"/1 \\\"%02x\\\"\" {}".format(
                " ".join([m[1] for m in modules])))

        if version_strings:
            for (m, f), v in zip(modules,
                                 version_strings.split(
                                     "040000001400000003000000474e5500")[1:]):
                module_versions[m] = v

    return module_versions


def get_wifi_data():
    wifi_data = {}

    unwanted_keys = ["Interface", "ifindex", "wdev", "wiphy"]

    output = get_command_output("iw dev")
    iface = None
    if output is not None:

        for line in output.splitlines():

            parts = line.split()
            if len(parts) < 2:
                continue

            k, v = parts[0], parts[1]

            if k == 'Interface':
                iface = v
                wifi_data[iface] = {}
                continue
            elif iface is None:
                continue

            if k in unwanted_keys:
                continue

            if k == 'txpower':
                v = float(v)

            if k == 'channel':
                # This condition will return a dict with all the values of the channel
                # With the input "channel 1 (2412 MHz), width: 20 MHz, center1: 2412 MHz"
                # the output will be {'addr':..., channel': {'band': 2462, 'center1': 2462, 'number': 11, 'width': 20}, 'ssid':...}
                v = {}
                v['number'] = int(parts[1])
                v['band'] = int(parts[2].strip("("))
                v['width'] = int(parts[5])
                v['center1'] = int(parts[8])

            if line.strip() == "multicast TXQ:":
                # No interesting output after this
                iface = None
                continue

            wifi_data[iface][k] = v

    return wifi_data



# -*- coding: utf-8 -*-
#
# testenv.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     17 September 2014
# Copyright (c) 2014-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import os

from collections import OrderedDict
from copy import deepcopy
from itertools import cycle, islice

from flent.util import Glob, token_split, parse_int
from flent.build_info import DATA_DIR
from flent.loggers import get_logger

TEST_PATH = os.path.join(DATA_DIR, 'tests')
logger = get_logger(__name__)

try:
    from os import cpu_count
except ImportError:
    from multiprocessing import cpu_count

try:
    CPU_COUNT = cpu_count()
except NotImplementedError:
    CPU_COUNT = 1

try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest

SPECIAL_PARAM_NAMES = ['upload_streams', 'download_streams', 'bidir_streams']
SPECIAL_PARAM_MAP = {'num_cpus': CPU_COUNT}

# Test parameters that will be parsed and from test parameters and passed to the
# add_stream callback. In addition to these, 'stream_delay' is handled specially
# in the code below
STREAM_CONFIG_PARAM_NAMES = ['label', 'ping_label', 'marking',
                             'control_host', 'local_bind', 'cc_algo',
                             'udp_bandwidth', 'udp_pktsize', 'send_size', 'test_payload']

# Mapping of test parameters that will be picked up from the global settings if
# they are not set
GLOBAL_TEST_PARAMS_MAP = {'local_binds': 'LOCAL_BIND',
                          'send_sizes': 'SEND_SIZE'}

class _no_default():
    pass


def finder(fn):
    """Decorator to put on find_* methods that makes sure common operations
    (i.e. skip if self.informational is set) are carried out correctly."""

    def decorated(self, *args, **kwargs):
        if self.informational:
            return ""
        ret = fn(self, *args, **kwargs)
        return ret
    return decorated


class TestEnvironment(object):

    def __init__(self, env={}, informational=False):
        self.env = self.replace_testparms(deepcopy(env))
        self.env.update({
            'glob': Glob,
            'o': OrderedDict,
            'include': self.include_test,
            'min_host_count': self.require_host_count,
            'set_test_parameter': self.set_test_parameter,
            'get_test_parameter': self.get_test_parameter,
            'try_test_parameters': self.try_test_parameters,
            'parse_int': self.parse_int,
            'zip_longest': zip_longest,
            'for_stream_config': self.for_stream_config,
            'test_error': self.test_error,
        })

        self.informational = informational
        self.orig_hosts = self.env['HOSTS']

    def execute(self, filename):
        try:
            with open(filename) as fp:
                exec(compile(fp.read(), filename, 'exec'), self.env)

            # Informational loading can override HOSTS to satisfy
            # require_host_count(); this should not be propagated.
            if self.informational:
                self.env['HOSTS'] = self.orig_hosts
            return self.expand_duplicates(self.env)
        except Exception as e:
            testn = os.path.basename(filename).replace(".conf", "")
            raise RuntimeError(
                "Error loading test '%s': %s." % (testn, e))

    def replace_testparms(self, env):
        if 'TEST_PARAMETERS' not in env:
            return env

        tp = env['TEST_PARAMETERS']
        for k in SPECIAL_PARAM_NAMES:
            if k in tp and tp[k] in SPECIAL_PARAM_MAP:
                tp[k] = SPECIAL_PARAM_MAP[tp[k]]
        return env

    def expand_duplicates(self, env):
        new_data_sets = []
        if 'DATA_SETS' not in env:
            return env
        for k, v in env['DATA_SETS'].items():
            try:
                for i in range(int(v['duplicates'])):
                    new_data_sets.append(
                        ("%s::%d" % (k, i + 1), dict(v,
                                                     id=str(i + 1),
                                                     duplicates=None)))
            except (KeyError, TypeError):
                new_data_sets.append((k, v))
            except ValueError:
                raise RuntimeError(
                    "Invalid number of duplicates: %s" % v['duplicates'])
            env['DATA_SETS'] = OrderedDict(new_data_sets)
        return env

    def include_test(self, name, env=None):
        self.execute(os.path.join(TEST_PATH, name))

    def set_test_parameter(self, name, value):
        self.env['TEST_PARAMETERS'][name] = value

    def get_test_parameter(self, name, default=_no_default, split=False, cast=None):
        try:
            ret = self.env['TEST_PARAMETERS'][name]
            if split:
                ret = token_split(ret)
                if cast:
                    ret = list(map(cast, ret))
            elif cast:
                ret = cast(ret)
            return ret
        except KeyError:
            if name in GLOBAL_TEST_PARAMS_MAP:
                ret = self.env[GLOBAL_TEST_PARAMS_MAP[name]]
                if ret:
                    return ret
            if default is not _no_default:
                return default
            if self.informational:
                return None
            raise RuntimeError("Missing required test parameter: %s" % name)

    def try_test_parameters(self, names, default=_no_default, split=False,
                            cast=None):
        name = names[0]
        for n in names:
            if n in self.env['TEST_PARAMETERS']:
                name = n
                break

        return self.get_test_parameter(name, default, split, cast)

    def parse_int(self, val):
        try:
            return parse_int(val)
        except ValueError:
            raise RuntimeError("Invalid integer value: %s" % val)

    def require_host_count(self, count):
        if len(self.env['HOSTS']) < count:
            if self.informational:
                self.env['HOSTS'] = ['dummy'] * count
            elif 'DEFAULTS' in self.env and 'HOSTS' in self.env['DEFAULTS'] \
                 and self.env['DEFAULTS']['HOSTS']:
                # If a default HOSTS list is set, populate the HOSTS list with
                # values from this list, repeating as necessary up to count
                def_hosts = cycle(self.env['DEFAULTS']['HOSTS'])
                missing_c = count - len(self.env['HOSTS'])
                self.env['HOSTS'].extend(islice(def_hosts, missing_c))
                if not self.env['HOST']:
                    self.env['HOST'] = self.env['HOSTS'][0]
            else:
                raise RuntimeError("Need %d hosts, only %d specified" %
                                   (count, len(self.env['HOSTS'])))

    def test_error(self, msg):
        if not self.informational:
            raise RuntimeError(msg)

    def for_stream_config(self, func, n=None):
        if n is None:
            n = len(self.env['HOSTS'])

        config_params = {}
        for k in STREAM_CONFIG_PARAM_NAMES:
            config_params[k] = self.get_test_parameter(k+"s",
                                                       default=[],
                                                       split=True)

        stream_delays = self.get_test_parameter("stream_delays",
                                                default=[],
                                                split=True)
        global_delay = self.env['DELAY']
        total_length = self.env['TOTAL_LENGTH']
        stream_length = total_length-2*global_delay

        for i in range(n):
            kwargs = {}
            try:
                kwargs['host'] = self.env['HOSTS'][i]
            except IndexError:
                pass

            for k, v in config_params.items():
                try:
                    kwargs[k] = v[i]
                except IndexError:
                    pass

            try:
                delay = int(stream_delays[i])
                kwargs['length'] = stream_length - delay
                kwargs['delay'] = global_delay + delay
            except IndexError:
                pass

            func(i, **kwargs)



# -*- coding: utf-8 -*-
#
# build_info.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:      6 December 2012
# Copyright (c) 2012-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# setup.py rewrites this file with the install prefix info

import os

# this value works for the source distribution
VERSION = "2.2.0"
DATA_DIR = os.path.dirname(__file__)

ENCODING = "UTF-8"
try:
    import locale
    ENCODING = locale.getpreferredencoding(False)
except:
    pass

if VERSION.endswith("+git") and os.path.exists(
        os.path.realpath(os.path.join(DATA_DIR, '..', '.git'))):
    try:
        import subprocess
        commit = subprocess.check_output(
            ["git", "log", "--format=%h", "--abbrev=8", "-1"], cwd=DATA_DIR).decode(ENCODING)
        VERSION += ".%s" % commit.strip()
    except:
        pass



# -*- coding: utf-8 -*-
#
# test_gui.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:      4 July 2019
# Copyright (c) 2019, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import unittest
import os
import sys

from .test_helpers import ForkingTestCase

from flent.settings import parser, Settings, DEFAULT_SETTINGS
settings = parser.parse_args(args=[], namespace=Settings(DEFAULT_SETTINGS))


class TestGui(ForkingTestCase):

    def setUp(self):
        self.settings = settings.copy()

        try:
            from qtpy import QtCore
        except ImportError:
            self.skipTest("No usable Qt module found")

        if (sys.platform == 'linux' and not os.getenv("DISPLAY")):
            self.skipTest("No DISPLAY variable set")

    def test_start_gui(self):
        from flent import gui
        gui.run_gui(self.settings, test_mode=True)


test_suite = unittest.TestSuite(
    [unittest.TestLoader().loadTestsFromTestCase(TestGui)])



# -*- coding: utf-8 -*-
#
# testcontrol-web-service.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     25 October 2014
# Copyright (c) 2014-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import json
import os
import re
import select
import shlex
import subprocess
import sys

try:
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
    from SocketServer import ForkingMixIn
except ImportError:
    from http.server import HTTPServer, BaseHTTPRequestHandler
    from socketserver import ForkingMixIn

tests = [{'name': 'test',
          'args': {'ipver': 'IP version (4 or 6)'},
          'exec': './test-exec.sh ${ipver}'}]


class TestWebServer(ForkingMixIn, HTTPServer):
    pass


class TestWebService(BaseHTTPRequestHandler):
    _INTERP_REGEX = re.compile(r"(^|[^$])(\$\{([^}]+)\})")
    _MAX_INTERP = 1000

    def respond(self, obj):
        self.send_response(200)
        self.send_header("Content-type", "application/json; charset=utf8")
        self.end_headers()
        self.wfile.write(json.dumps(obj).encode("utf-8"))
        self.wfile.write("\n".encode("utf-8"))

    def find_test(self, name):
        for t in tests:
            if t['name'] == name:
                return t
        self.send_error(404, "Test not found: %s" % name)
        return None

    def do_GET(self):
        print(self.headers)
        parts = os.path.split(self.path)
        if len(parts) < 3 and not parts[1]:
            self.respond(list([i['name'] for i in tests]))
        else:
            test = self.find_test(parts[1])
            if test:
                self.respond(test['args'])

    def parse_input(self):
        length = self.headers.get('content-length')
        input_type = self.headers.get('Content-Type')
        parts = input_type.split(";", 1)
        if parts[0].strip().lower() != 'application/json':
            self.send_error(415, 'Invalid format. Needs application/json')
            return None
        if len(parts) > 1 and parts[1].startswith('charset='):
            charset = parts[1].split("=", 1)[1].strip()
        else:
            charset = 'utf-8'

        try:
            nbytes = int(length)
        except (TypeError, ValueError):
            self.send_error(411)
            return None

        if nbytes > 0:
            data = self.rfile.read(nbytes)

        while select.select([self.rfile], [], [], 0)[0]:
            if not self.rfile.read(1):
                break

        try:
            return json.loads(data.decode(charset))
        except Exception as e:
            self.send_error(400, "Unable to decode: %s" % e)

    def interpolate(self, string, req):
        m = self._INTERP_REGEX.search(string)
        i = 0
        while m is not None:
            k = m.group(3)
            if k in req:
                string = string.replace(m.group(2), str(req[k]))
            else:
                string = string.replace(m.group(2), "$${%s}" % k)
            m = self._INTERP_REGEX.search(string)
            i += 1
            if i > self._MAX_INTERP:
                self.send_error(
                    500, "Too many interpolations performed for exec string.")
                return None
        return string.replace("$$", "$")

    def handle_test(self, test, req):
        cmdline = self.interpolate(test['exec'], req)
        if cmdline:
            args = shlex.split(cmdline)
            proc = subprocess.Popen(args, stdout=subprocess.PIPE)
            resp = json.loads(proc.stdout.readline().decode('utf-8'))
            self.respond(resp)

    def do_POST(self):
        parts = os.path.split(self.path)
        if not parts[1]:
            self.send_error(405)
            return
        test = self.find_test(parts[1])
        if test:
            req = self.parse_input()
            if req:
                self.handle_test(test, req)


if __name__ == "__main__":
    server_addr = ('localhost', 8000)
    server = TestWebServer(server_addr, TestWebService)
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        sys.exit(0)



# -*- coding: utf-8 -*-
#
# settings.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     25 November 2012
# Copyright (c) 2012-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import os
import socket
import sys

from datetime import datetime
from copy import deepcopy
from collections import OrderedDict

try:
    from configparser import RawConfigParser
except ImportError:
    from ConfigParser import RawConfigParser

from flent.build_info import VERSION
from flent.testenv import TestEnvironment, TEST_PATH
from flent.loggers import get_logger
from flent.util import FuncAction, Update, AddHost, append_host, keyval, \
    keyval_int, keyval_transformer, ArgParser, token_split, global_cache, \
    utcnow
from flent.plotters import add_plotting_args
from flent import loggers, util, resultset, runners

# Python 2/3 compatibility
try:
    unicode
except NameError:
    unicode = str

logger = get_logger(__name__)

OLD_RCFILE = os.path.expanduser("~/.netperf-wrapperrc")

DEFAULT_SETTINGS = {
    'NAME': None,
    'HOST': None,
    'LOCAL_HOST': socket.gethostname(),
    'DESCRIPTION': 'No description',
    'PLOTS': {},
    'TIME': utcnow(),
    'BATCH_NAME': None,
    'BATCH_UUID': None,
}

DICT_SETTINGS = ('DATA_SETS', 'PLOTS')


class _LOG_DEFER:
    pass


class Version(FuncAction):

    def __call__(*args):
        logger.info("Flent v%s.\nRunning on Python %s.",
                    VERSION, sys.version.replace("\n", " "))
        try:
            import matplotlib
            import numpy
            logger.info("Using matplotlib version %s on numpy %s.",
                        matplotlib.__version__, numpy.__version__)
        except ImportError:
            logger.info("No matplotlib found. Plots won't be available.")
        try:
            import qtpy
            from qtpy import QtCore
            logger.info("Using Qt: %s v%s.", qtpy.API, QtCore.__version__)
        except ImportError:
            logger.info("No usable Qt version found. GUI won't work.")
        sys.exit(0)


class LogLevel(FuncAction):

    def __init__(self, option_strings, dest, level=None, **kwargs):
        super(LogLevel, self).__init__(option_strings, dest, **kwargs)
        self.level = level

    def __call__(self, parser, namespace, values, option_string=None):
        loggers.set_console_level(self.level)
        setattr(namespace, self.dest, self.level)


class LogFile(argparse.Action):

    def __init__(self, option_strings, dest, level=None, **kwargs):
        super(LogFile, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if values is None:
            setattr(namespace, self.dest, _LOG_DEFER)
        else:
            loggers.setup_logfile(values)
            setattr(namespace, self.dest, values)


class Debug(argparse.Action):
    def __init__(self, option_strings, dest, help=None):
        super(Debug, self).__init__(option_strings, dest,
                                    default=False, nargs=0,
                                    help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        loggers.enable_exceptions()
        setattr(namespace, self.dest, True)


class ListTests(FuncAction):

    @classmethod
    def get_tests(cls, settings):
        settings = settings.copy()
        tests = []
        for t in sorted([os.path.splitext(i)[0] for i in os.listdir(TEST_PATH)
                         if i.endswith('.conf')]):
            try:
                settings.load_test(t, informational=True)
                tests.append((t, settings.DESCRIPTION))
            except Exception as e:
                logger.exception(str(e))
        return tests

    def __call__(self, parser, namespace, values, option_string=None):
        tests = self.get_tests(namespace)
        logger.info('Available tests:')
        max_len = max([len(t[0]) for t in tests])
        for t, desc in tests:
            desc = desc.replace("\n", "\n" + " " * (max_len + 6))
            logger.info("  %-" + str(max_len) + "s :  %s", t, desc)
        sys.exit(0)


parser = ArgParser(description='The FLExible Network Tester.')

parser.add_argument(
    "args", nargs="*", type=unicode, metavar="host|test|input_file",
    help="Hostname, test name or input filenames.")

parser.add_argument(
    "-o", "--output",
    action="store", type=unicode, dest="OUTPUT", default="-",
    help="File to write processed output to (default standard out).")

parser.add_argument(
    "-D", "--data-dir",
    action="store", type=unicode, dest="DATA_DIR",
    help="Directory to store data files in. Defaults to the current directory.")

parser.add_argument(
    "-i", "--input",
    action="append", type=unicode, dest="INPUT", default=[],
    help="File to read input from (instead of running tests). "
    "Input files can also be specified as unqualified arguments "
    "without using the -i switch.")

parser.add_argument(
    "-f", "--format",
    action="store", type=unicode, dest="FORMAT", default='default',
    help="Select output format (one of plot, csv, org_table, stats, "
    "metadata, summary). Default: 'summary'.")

parser.add_argument(
    "-p", "--plot",
    action="store", type=unicode, dest="PLOT",
    help="Select which plot to output for the given test (implies -f plot if no"
    "formatter is selected). Use the --list-plots option to see available plots.")

parser.add_argument(
    "-t", "--title-extra",
    action="store", type=unicode, dest="TITLE", default="",
    help="Text to add to plot title and data file name.")

parser.add_argument(
    "-n", "--note",
    action="store", type=unicode, dest="NOTE",
    help="Add arbitrary text as a note to be stored in the JSON data file "
    "(under the NOTE key in the metadata object).")

parser.add_argument(
    "-r", "--rcfile",
    action="store", type=unicode, dest="RCFILE",
    default=os.path.expanduser("~/.flentrc"),
    help="Load configuration data from RCFILE (default ~/.flentrc).")

parser.add_argument(
    "-x", "--extended-metadata",
    action="store_true", dest="EXTENDED_METADATA",
    help="Collect extended metadata and store it with the data file. "
    "May include details of your machine you don't want to distribute; "
    "see man page.")

parser.add_argument(
    "--remote-metadata",
    action="append", type=unicode, dest="REMOTE_METADATA",
    metavar="HOSTNAME", default=[],
    help="Collect extended metadata from a remote host. HOSTNAME is passed "
    "verbatim to ssh, so can include hosts specified in ~/.ssh/config. This "
    "option can be specified multiple times. Note that gathering the data can "
    "take some time, since it involves executing several remote commands. "
    "This option implies --extended-metadata.")

parser.add_argument(
    "--gui",
    action="store_true", dest="GUI",
    help="Run the Flent GUI. All other options are used as defaults "
    "in the GUI, but can be changed once it is running.")

parser.add_argument(
    "--new-gui-instance",
    action="store_true", dest="NEW_GUI_INSTANCE",
    help="Start a new GUI instance. Otherwise, flent will try to "
    "connect to an already running GUI instance and have that load any new "
    "data files specified as arguments. Implies --gui.")

parser.add_argument(
    "--gui-no-defer", action="store_true", dest="GUI_NO_DEFER",
    help="Normally, the GUI defers redrawing plots until they are needed to "
    "avoid redrawing all open plots every time an option changes. This switch "
    "turns off that optimisation in favour of always redrawing everything "
    "straight away. This is useful when loading a bunch of plots from the "
    "command line and then wanting to flip through them without drawing delay.")

parser.add_argument(
    "-b", "--batch",
    action="append", type=unicode, dest="BATCH_NAMES", metavar="BATCH_NAME",
    default=[], help="Run test batch BATCH_NAME (must be specified in a batch "
    "file loaded by the --batch-file option). Can be supplied multiple times.")

parser.add_argument(
    "-B", "--batch-file", action="append", type=unicode, dest="BATCH_FILES",
    metavar="BATCH_FILE", default=[],
    help="Load batch file BATCH_FILE. Can be specified multiple times, in which "
    "case the files will be combined (with identically-named sections being "
    "overridden by later files). See the man page for an explanation of the "
    "batch file format.")

parser.add_argument(
    "--batch-override",
    action=Update, type=keyval, dest="BATCH_OVERRIDE", metavar="key=value",
    help="Override parameter 'key' in the batch config and set it to 'value'. "
    "The key name will be case folded to lower case. Can be specified multiple "
    "times.")

parser.add_argument(
    "--batch-dry-run",
    action="store_true", dest="BATCH_DRY",
    help="Dry batch run. Prints what would be done, but doesn't actually run any "
    "tests.")

parser.add_argument(
    "--batch-verbose",
    action="store_true", dest="BATCH_VERBOSE",
    help="Be verbose during batch run: Print all commands executed.")

parser.add_argument(
    "--batch-no-shuffle",
    action="store_false", dest="BATCH_SHUFFLE",
    help="Do not randomise the order of test runs within each batch.")

parser.add_argument(
    "--batch-no-timestamp",
    action="store_false", dest="BATCH_TIMESTAMP",
    help="Do not add a timestamp to batch output filenames. This keeps filenames "
    "stable, but requires that the filename_extra variable be unique for each "
    "test run. Results will be overwritten and a warning emitted if not.")

parser.add_argument(
    "--batch-repetitions", action="store", type=int, dest="BATCH_REPS",
    metavar="REPETITIONS",
    help="Shorthand for --batch-override 'repetitions=REPETITIONS'.")

parser.add_argument(
    "--batch-title",
    action="store", type=unicode, dest="BATCH_TITLE", metavar="TITLE",
    help="Shorthand for --batch-override 'batch_title=TITLE'.")

parser.add_argument(
    "--batch-resume",
    action="store", type=unicode, dest="BATCH_RESUME", metavar="DIR",
    help="Try to resume a previously interrupted batch run. The argument is "
    "the top-level output directory from the previous run. Tests for which "
    "data files already exist will be skipped.")

test_group = parser.add_argument_group(
    "Test configuration",
    "These options affect the behaviour of the test being run "
    "and have no effect when parsing input files.")

test_group.add_argument(
    "-H", "--host",
    action=AddHost, type=unicode, dest="HOSTS", metavar='HOST', default=[],
    help="Host to connect to for tests. For tests that support it, multiple "
    "hosts can be specified by supplying this option multiple times. Hosts can "
    "also be specified as unqualified arguments; this parameter guarantees that "
    "the argument be interpreted as a host name (rather than being subject to "
    "auto-detection between input files, hostnames and test names).")

test_group.add_argument(
    "--local-bind",
    action="append", type=unicode, dest="LOCAL_BIND", metavar='IP', default=[],
    help="Local hostname or IP address to bind to (for test tools that support "
    "this). Can be specified multiple times to get different local bind address "
    "per host.")

test_group.add_argument(
    "--remote-host",
    action=Update, type=keyval_int, dest="REMOTE_HOSTS", metavar='idx=HOSTNAME',
    help="A remote hostname to connect to when starting a test. The idx is "
    "the runner index, which is assigned sequentially by the number of *runners* "
    "(which is *not* the same as the number of hosts). Look for the 'IDX' key in "
    "SERIES_META for a test get the idx used here. This works by simply "
    "prepending 'ssh HOSTNAME' to the runner command, so it relies on the same "
    "binaries being in the same places on both machines, and won't work for all "
    "runners. Can be specified multiple times.")

test_group.add_argument(
    "-l", "--length",
    action="store", type=int, dest="LENGTH", default=60,
    help="Base test length (some tests may add some time to this).")

test_group.add_argument(
    "-s", "--step-size",
    action="store", type=float, dest="STEP_SIZE", default=0.2,
    help="Measurement data point step size.")

test_group.add_argument(
    "-d", "--delay",
    action="store", type=int, dest="DELAY", default=5,
    help="Number of seconds to delay parts of test (such as bandwidth loaders).")

test_group.add_argument(
    "-4", "--ipv4",
    action="store_const", const=4, dest="IP_VERSION",
    help="Use IPv4 for tests (some tests may ignore this).")

test_group.add_argument(
    "-6", "--ipv6",
    action="store_const", const=6, dest="IP_VERSION",
    help="Use IPv6 for tests (some tests may ignore this).")

test_group.add_argument(
    "--socket-timeout",
    action="store", type=int, dest="SOCKET_TIMEOUT", default=2,
    help="Socket timeout (in seconds) used for UDP delay measurement, to prevent "
    "stalls on packet loss. Only enabled if the installed netperf version is "
    "detected to support this (requires netperf version 2.7 or newer). Set to 0 "
    "to disable.")

test_group.add_argument(
    "--send-size",
    action="append", type=unicode, dest="SEND_SIZE", default=[],
    help="Send size (in bytes) used for TCP tests. netperf uses the socket "
    "buffer size by default, which if too large can cause spikes in the "
    "throughput results. Lowering this value will increase CPU usage but "
    "also improves the fidelity of the throughput results without having "
    "to decrease the socket buffer size. Can be specified multiple times, "
    "with each value corresponding to a stream of a test.")

test_group.add_argument(
    "--test-payload",
    action="store", type=unicode, dest="TEST_PAYLOAD", default='/dev/urandom',
    help="Path to file containing payload to pre-fill the netperf buffers with "
    "defaults to randomised payload")

test_group.add_argument(
    "--test-parameter",
    action=Update, type=keyval, dest="TEST_PARAMETERS", metavar='key=value',
    help="Arbitrary test parameter in key=value format. "
    "Key will be case folded to lower case. Some test configurations may "
    "alter behaviour based on values passed as test parameters. Additionally, "
    "the values are stored with the results metadata, and so can be used for "
    "arbitrary resultset categorisation. Can be specified multiple times.")

test_group.add_argument(
    "--swap-up-down",
    action="store_true", dest="SWAP_UPDOWN",
    help="Swap upstream and downstream directions for data transfer. This means "
    "that 'upload' will become 'download' and vice versa. Works by exchanging "
    "netperf TCP_MAERTS and TCP_STREAM parameters, so only works for tests "
    "that employ these as their data transfer, and only for the TCP streams.")

test_group.add_argument(
    "--socket-stats",
    action="store_true", dest="SOCKET_STATS",
    help="Parse socket stats during test. This will capture and parse socket "
    "statistics for all TCP upload flows during a test, adding TCP cwnd and RTT "
    "values to the test data. Requires the 'ss' utility to be present on the "
    "system, and can fail if there are too many simultaneous upload flows; which "
    "is why this option is not enabled by default.")

test_group.add_argument(
    "--marking-name",
    action=Update, dest="MARKING_NAMES", metavar='name=hexcode',
    type=keyval_transformer(keyfunc=lambda x: x.upper(),
                            valfunc=util.parse_int,
                            errmsg="Values must be integers"),
    help="Define a new symbolic name that can be used when specifying flow "
    "markings using the 'markings' test parameter. This can be used to make "
    "it easier to specify custom diffserv markings on flows by using symbolic "
    "names for each marking value instead of the hex codes. Values specified "
    "here will be used in addition to the common values (AFxx, CSx and EF - "
    "see man page for full list), and cannot override the built-in names. "
    "Names will be case-folded when matching.")

plot_group = parser.add_argument_group(
    "Plot configuration",
    "These options are used to configure the appearance of "
    "plot output and only make sense combined with -f plot.")

add_plotting_args(plot_group)

tool_group = parser.add_argument_group("Test tool-related options")

tool_group.add_argument(
    "--control-host",
    action="store", type=unicode, dest="CONTROL_HOST", metavar="HOST",
    help="Hostname for control connection for test tools that support it "
    "(netperf and D_ITG). If not supplied, this will be the same as the test "
    "target host. The per-flow test parameter setting takes precedence of this "
    "for multi-target tests.")

tool_group.add_argument(
    "--control-local-bind",
    action="store", type=unicode, dest="CONTROL_LOCAL_BIND", metavar="IP",
    help="Local IP to bind control connection to (for test tools that support it;"
    " currently netperf). If not supplied, the value for --local-bind will be "
    "used.")

tool_group.add_argument(
    "--netperf-control-port",
    action="store", type=int, dest="NETPERF_CONTROL_PORT", metavar="PORT",
    default=12865, help="Port for Netperf control server.")

tool_group.add_argument(
    "--ditg-control-port",
    action="store", type=int, dest="DITG_CONTROL_PORT", metavar="PORT",
    default=8000, help="Port for D-ITG control server.")

tool_group.add_argument(
    "--ditg-control-secret",
    action="store", type=unicode, dest="DITG_CONTROL_SECRET", metavar="SECRET",
    default='', help="Secret for D-ITG control server authentication.")

tool_group.add_argument(
    "--http-getter-urllist",
    action="append", type=unicode, dest="HTTP_GETTER_URLLIST", metavar="FILENAME",
    help="Filename containing the list of HTTP URLs to get. Can also be a URL, "
    "which will then be downloaded as part of each test iteration. If not "
    "specified, this is set to http://<hostname>/filelist.txt where <hostname> "
    "is the first test hostname. Can be specified multiple times, in which case "
    "one http-getter instance will be run for each URL list.")

tool_group.add_argument(
    "--http-getter-dns-servers",
    action="store", type=unicode, dest="HTTP_GETTER_DNS", metavar="DNS_SERVERS",
    help="DNS servers to use for http-getter lookups. Format is "
    "host[:port][,host[:port]]... This option will only work if libcurl supports "
    "it (needs to be built with the ares resolver). "
    "Default is none (use the system resolver).")

tool_group.add_argument(
    "--http-getter-timeout",
    action="store", type=int, dest="HTTP_GETTER_TIMEOUT", metavar="MILLISECONDS",
    help="Timeout for HTTP connections. Default is to use the test length.")

tool_group.add_argument(
    "--http-getter-workers",
    action="store", type=int, dest="HTTP_GETTER_WORKERS", metavar="NUMBER",
    default=4, help="Number of workers to use for getting HTTP urls. "
    "Default is 4.")

tool_group.add_argument(
    "--irtt-sampling-interval",
    action="store", type=int, dest="IRTT_INTERVAL", metavar="MILLISECONDS",
    help="Override the sampling interval passed to irtt, in milliseconds. Can be "
    "used to run irtt with a higher sampling frequency than the rest of the test. "
    "If set, this will override the sampling interval for all instances of irtt "
    "used in the test.")

misc_group = parser.add_argument_group("Misc and debugging options")

misc_group.add_argument(
    "-L", "--log-file",
    action=LogFile, type=unicode, dest="LOG_FILE", nargs='?',
    help="Write debug log (test program output) to log file. If the option is "
    "enabled but no file name is given, the log file name is derived from the "
    "test data filename.")

misc_group.add_argument(
    "--cache-file",
    action="store", type=unicode, dest="CACHE_FILE", metavar="FILENAME",
    help="Use this file for caching lookup of runner file names and "
    "versions between runs.")

misc_group.add_argument(
    '--list-tests',
    action=ListTests,
    help="List available tests and exit.")

misc_group.add_argument(
    '--list-plots',
    action='store_true', dest="LIST_PLOTS",
    help="List available plots for selected test and exit.")

misc_group.add_argument(
    "-V", "--version",
    action=Version,
    help="Show flent version information and exit.")

misc_group.add_argument(
    "-v", "--verbose",
    action=LogLevel, level=loggers.DEBUG, dest="LOG_LEVEL",
    help="Enable verbose logging to console.")

misc_group.add_argument(
    "-q", "--quiet",
    action=LogLevel, level=loggers.WARNING, dest="LOG_LEVEL",
    help="Disable normal logging to console (and only log warnings and errors).")

misc_group.add_argument(
    "--debug-error",
    action=Debug, dest="DEBUG_ERROR",
    help="Print full exception backtraces to console.")


def new():
    return parser.parse_args([], namespace=Settings(DEFAULT_SETTINGS))


class Settings(argparse.Namespace):

    FLENT_VERSION = VERSION

    def __init__(self, defs):

        # Copy everything from defaults to make sure the defaults are not
        # modified.
        defaults = {}
        for k, v in defs.items():
            defaults[k] = deepcopy(v)
        argparse.Namespace.__init__(self, **defaults)

    def load_test_or_host(self, test_name):
        filename = os.path.join(TEST_PATH, test_name + ".conf")

        if not os.path.exists(filename):
            # Test not found, assume it's a hostname
            append_host(self.HOSTS, test_name)
        elif self.NAME is not None and self.NAME != test_name:
            raise RuntimeError("Multiple test names specified.")
        else:
            self.NAME = test_name

    def load_rcfile(self):
        self.process_args()
        if self.RCFILE == parser.get_default('RCFILE') and \
           not os.path.exists(self.RCFILE) and os.path.exists(OLD_RCFILE):
            logger.warning("Using old rcfile found at %s, "
                           "please rename to %s.",
                           OLD_RCFILE, self.RCFILE)
            self.RCFILE = OLD_RCFILE
        if os.path.exists(self.RCFILE):
            logger.debug("Loading rc file %s", self.RCFILE)

            config = RawConfigParser()
            config.optionxform = lambda x: x.upper()
            config.read(self.RCFILE)

            items = []

            if config.has_section('global'):
                items.extend(config.items('global'))
            if self.NAME is not None and config.has_section(self.NAME):
                items.extend(config.items(self.NAME))
            try:
                return self.parse_rcvalues(items)
            except (ValueError, argparse.ArgumentTypeError) as e:
                raise RuntimeError("Unable to parse RC values: %s" % e)
        return {}

    def parse_rcvalues(self, items):

        vals = {}

        for k, v in items:
            k = k.upper()
            t = parser.get_type(k)
            if t == bool:
                if type(v) == bool:
                    vals[k] = v
                elif v.lower() in ('1', 'yes', 'true', 'on'):
                    vals[k] = True
                elif v.lower() in ('0', 'no', 'false', 'off'):
                    vals[k] = False
                else:
                    raise ValueError("Not a boolean: %s" % v)
                logger.debug("Set value %s=%s from rc file", k, vals[k])
                continue

            elif t:

                val = t(v)
                c = parser.get_choices(k)
                if c and val not in c:
                    logger.warning("Invalid RC value '%s' for key %s. Ignoring",
                                   val, k)
                    continue
                if isinstance(val, dict) and k in vals:
                    vals[k].update(val)
                elif parser.is_list(k):
                    vals[k] = [t(i.strip()) for i in token_split(v)]
                else:
                    vals[k] = val
                logger.debug("Set value %s=%s from rc file", k, vals[k])

        return vals

    def load_rcvalues(self, vals, override=False):

        try:
            vals = self.parse_rcvalues(vals)
        except (ValueError, argparse.ArgumentTypeError) as e:
            raise RuntimeError("Unable to parse RC values: %s" % e)

        for k, v in vals.items():
            if override or getattr(self, k) == parser.get_default(k):
                setattr(self, k, v)

        self.update_implications()

    def load_test(self, test_name=None, informational=False):
        if test_name is not None:
            self.NAME = test_name
        if self.HOSTS:
            self.HOST = self.HOSTS[0]
        if hasattr(self, 'TOTAL_LENGTH'):
            self.TOTAL_LENGTH = self.LENGTH

        if not informational:
            self.lookup_hosts()

        if self.NAME is None:
            if informational:
                # Informational lookups should not fail
                return
            raise RuntimeError("Missing test name.")
        test_env = TestEnvironment(self.__dict__, informational)
        filename = os.path.join(TEST_PATH, self.NAME + ".conf")
        logger.debug("Executing test environment file %s", filename)
        s = test_env.execute(filename)

        for k, v in list(s.items()):
            if k == k.upper():
                setattr(self, k, v)

        self.update_defaults()

    def update_defaults(self):
        if hasattr(self, 'DEFAULTS'):
            for k, v in list(self.DEFAULTS.items()):
                if not hasattr(self, k) or getattr(self, k) is None:
                    setattr(self, k, v)

    def compute_missing_results(self, results):
        if "FROM_COMBINER" in results.meta():
            return
        for dname, dvals in self.DATA_SETS.items():
            if dname not in results:
                runner = runners.get(dvals['runner'])
                if issubclass(runner, runners.ComputingRunner):
                    logger.debug("Computing missing data series %s", dname)
                    try:
                        runner = runner(name=dname, settings=self,
                                        post=True, **dvals)
                        runner.compute_result(results)
                    except Exception as e:
                        logger.exception("Unable to compute missing data "
                                         "series '%s': '%s'.", dname, e)
                        raise

    def lookup_hosts(self):
        """If no explicit IP version is set, do a hostname lookup and try to"""
        version = 4
        for h in self.HOSTS:
            try:
                hostname = util.lookup_host(h)
                if hostname[0] == socket.AF_INET6:
                    version = 6
            except socket.gaierror as e:
                raise RuntimeError("Hostname lookup failed for host %s: %s"
                                   % (h, e))

        if self.IP_VERSION is None:
            self.IP_VERSION = version

    def __setattr__(self, k, v):
        if k in DICT_SETTINGS and isinstance(v, list):
            v = OrderedDict(v)

        object.__setattr__(self, k, v)

    def update(self, values):
        updated = False
        for k, v in list(values.items()):
            if not hasattr(self, k) or getattr(self, k) != v:
                updated = True
                setattr(self, k, v)
        return updated

    def items(self):
        return self.__dict__.items()

    def copy(self):
        return Settings(self)

    def process_args(self):
        for v in 'INPUT', 'SCALE_DATA', 'HOSTS':
            if not getattr(self, v):
                setattr(self, v, [])
        while self.args:
            a = self.args.pop(0)
            if os.path.exists(a):
                if self.SCALE_MODE and self.INPUT:
                    self.SCALE_DATA.append(a)
                else:
                    self.INPUT.append(a)
            else:
                self.load_test_or_host(a)

    def update_implications(self):
        # If run with no args and no controlling TTY, launch the GUI by default
        if not sys.stdin.isatty() and not sys.stdout.isatty() and \
           not sys.stderr.isatty() and len(sys.argv) < 2:
            self.GUI = True
        # Passing --new-gui-instance on the command line implies --gui, but
        # setting it in the rc file does not. When set here, before the rc file
        # is loaded, this has the desired effect.
        elif self.NEW_GUI_INSTANCE:
            self.GUI = True

        if self.CACHE_FILE:
            global_cache.read_file(self.CACHE_FILE)

        if self.REMOTE_METADATA:
            self.EXTENDED_METADATA = True

        if self.PLOT is not None and self.FORMAT == 'default':
            self.FORMAT = 'plot'

        if self.BATCH_REPS is not None:
            self.BATCH_OVERRIDE['repetitions'] = self.BATCH_REPS

        if self.BATCH_TITLE is not None:
            self.BATCH_OVERRIDE['batch_title'] = self.BATCH_TITLE

        if self.HOST is None and self.HOSTS:
            self.HOST = self.HOSTS[0]

        if self.DATA_DIR is None:
            self.DATA_DIR = os.path.dirname(self.OUTPUT) or '.'

        # For per-stream options, if only specified once duplicate them for each
        # host so they apply to all flows
        if len(self.LOCAL_BIND) == 1 and len(self.HOSTS) > 1:
            self.LOCAL_BIND *= len(self.HOSTS)
        if len(self.SEND_SIZE) == 1 and len(self.HOSTS) > 1:
            self.SEND_SIZE *= len(self.HOSTS)

        for k, v in self.BATCH_OVERRIDE.items():
            if not hasattr(v, 'lower'):
                continue
            if v.lower() in ('no', 'false', '0'):
                self.BATCH_OVERRIDE[k] = False
            elif v.lower() in ('yes', 'true', '0'):
                self.BATCH_OVERRIDE[k] = True


def load_gui(settings):
    from flent import gui
    gui.run_gui(settings)  # does not return


def load(argv):
    logger.info("Starting Flent %s using Python %s.", VERSION,
                sys.version.split()[0])

    # We parse the args twice - the first pass is just to get the test name and
    # the name of the rcfile to parse in order to get the defaults
    settings = parser.parse_args(argv, namespace=Settings(DEFAULT_SETTINGS))
    parser.set_defaults(**{k: v for k, v in settings.load_rcfile().items()
                           if getattr(settings, k) == parser.get_default(k)})

    settings = parser.parse_args(argv, namespace=Settings(DEFAULT_SETTINGS))
    settings.process_args()
    settings.update_implications()

    if settings.SCALE_DATA:
        scale_data = []
        for filename in settings.SCALE_DATA:
            if filename in settings.INPUT:
                # Do not load input file twice - makes it easier to select a set
                # of files for plot scaling and supply each one to -i without
                # having to change the other command line options each time.
                continue
            r = resultset.load(filename)
            scale_data.append(r)
        settings.SCALE_DATA = scale_data

    settings.load_test(informational=True)

    if settings.LIST_PLOTS:
        list_plots(settings)

    return settings


def list_plots(settings):
    plots = list(settings.PLOTS.keys())
    if not plots:
        logger.error("No plots available for test '%s'.", settings.NAME)
        sys.exit(0)

    logger.info("Available plots for test '%s':", settings.NAME)
    max_len = str(max([len(p) for p in plots]))
    for p in plots:
        logger.info("  %-" + max_len + "s :  %s",
                    p, settings.PLOTS[p]['description'])
    sys.exit(0)



# -*- coding: utf-8 -*-
#
# ditg-control-server.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     28 marts 2014
# Copyright (c) 2014-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import calendar
import hashlib
import hmac
import json
import optparse
import os
import random
import shutil
import signal
import string
import subprocess
import sys
import tempfile
import time
import traceback

from datetime import datetime

try:
    from defusedxml.xmlrpc import monkey_patch
    monkey_patch()
    del monkey_patch
    XML_DEFUSED = True
except ImportError:
    XML_DEFUSED = False

try:
    # python 2
    from SimpleXMLRPCServer import SimpleXMLRPCServer
except ImportError:
    from xmlrpc.server import SimpleXMLRPCServer

# Nicked from the subprocess module -- for closing open file descriptors
try:
    MAXFD = os.sysconf("SC_OPEN_MAX")
except:
    MAXFD = 256

ALPHABET = list(string.ascii_lowercase) + list(string.digits)


class AlarmException(Exception):
    pass


parser = optparse.OptionParser(
    description="Control server for D-ITG receive component session negotiation.")

parser.add_option('--insecure-xml', action='store_true', dest='INSECURE_XML',
                  default=False,
                  help="Run even though the defusedxml module is unavailable. "
                  "WARNING: Will leave the server open to entity expansion "
                  "attacks!")
parser.add_option('-a', '--address', action='store', type='string',
                  dest='BIND_ADDRESS', default='localhost',
                  help="Address to bind to. Default: localhost.")
parser.add_option('-p', '--port', action='store', type='int', dest='BIND_PORT',
                  default='8000', help="Bind port. Default: 8000.")
parser.add_option('-A', '--itg-address', action='store', type='string',
                  dest='ITG_ADDRESS', default=None,
                  help="Address to bind ITGRecv to. Default: Same as --address.")
parser.add_option('-s', '--start-port', action='store', type='int',
                  dest='START_PORT', default=9000, help="Start port for ITGRecv "
                  "control socket binds (default 9000).")
parser.add_option('-t', '--max-test-time', action='store', type='int',
                  dest='MAX_TEST_TIME', default=7200, help="Maximum test time "
                  "allowed. Default 7200 seconds (two hours).")
parser.add_option('-m', '--max-instances', action='store', type='int',
                  dest='MAX_INSTANCES', default=100,
                  help="Maximum number of running instances before new requests "
                  "are denied (default 100).")
parser.add_option('-S', '--secret', action='store', type='string', dest='SECRET',
                  default="", help="Secret for request authentication. "
                  "Default: ''.")
parser.add_option('--debug', action='store_true', dest='DEBUG',
                  default=False, help="Debug mode: Don't delete temporary files.")


class DITGManager(object):
    datafile_pattern = "%s.json"

    def __init__(self, bind_address, start_port, max_test_time, max_instances,
                 secret, debug=False):
        self.working_dir = tempfile.mkdtemp(prefix='ditgman-')
        self.seen = {}
        self.children = []
        self.toplevel = True
        self.bind_address = bind_address
        self.max_test_time = max_test_time
        self.max_instances = max_instances
        self.hmac = hmac.new(secret.encode(), digestmod=hashlib.sha256)
        self.start_port = self.current_port = start_port
        self.id_length = 20
        self.debug = debug

        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)
        signal.signal(signal.SIGALRM, self._alarm)
        signal.signal(signal.SIGCHLD, self._collect_garbage)

    def get_test_results(self, test_id):
        """Get the results of a previously initiated test. The sole parameter is
        the test_id returned from request_new_test().

        The return parameter is a dictionary with the following entries:

        status: 'OK' if everything went well, 'Error' otherwise.

        message: Set if status is 'Error'; contains an error message.

        data: The output of ITGDec -c with the interval requested at initiation.

        utc_offset: UTC timestamp which the time values in the return data is
        offset by.

        """
        self._collect_garbage()
        test_id = str(test_id)
        if len(test_id) != self.id_length:
            return {'status': 'Error',
                    'message': "Invalid test id: '%s'." % test_id}
        for c in test_id:
            if c not in ALPHABET:
                return {'status': 'Error',
                        'message': "Invalid test id: '%s'." % test_id}
        filename = os.path.join(
            self.working_dir, self.datafile_pattern % test_id)
        if not os.path.exists(filename):
            return {'status': 'Error',
                    'message': "Data for test ID '%s' not found." % test_id}
        with open(filename, 'rt') as fp:
            data = json.load(fp)
            return data

    def request_new_test(self, duration, interval, hmac_hex, raw_data=False):
        """Request a new test instance.

        This will allocate a new ITGRecv instance and parse the log file it
        produces. The results of this can then be retrieved with
        get_test_results() after the test has run.

        The parameters are:

        duration: Requested test duration in seconds. The ITGRecv instance will
                  be killed after this time has passed (+ a grace period of five
                  seconds).

        interval: The requested interval for data points, in milliseconds
                  (passed to ITGDec).

        hmac_hex: A hexadecimal HMAC-SHA256 of the two other parameters computed
                  by concatenating their ASCII representations. The HMAC secret
                  is configured by the operator of the control server instance.

        raw_data: Whether to store and return the raw text log from ITGDec (i.e.
                  the output of ITGDec -l)

        The return value is a dictionary with the following keys:

        status: 'OK' if everything went well, 'Error' otherwise.

        message: Set if status is 'Error'; contains an error message.

        test_id: The assigned test ID, to be passed to get_test_results() after
                 the duration has expired.

        port: The control server port of the ITGRecv instance. The sender is
              expected to use port+1 for the data connection.

        """
        self._collect_garbage()
        duration = int(duration)
        interval = int(interval)
        hmac = self.hmac.copy()
        hmac.update(str(duration).encode())
        hmac.update(str(interval).encode())
        if hmac.hexdigest() != hmac_hex:
            return {'status': 'Error', 'message': "HMAC authentication failure."}
        if duration <= 0 or interval <= 0:
            return {'status': 'Error',
                    'message': "Duration and interval must be positive integers."}
        if duration > self.max_test_time:
            return {'status': 'Error',
                    'message': "Maximum test time of %d seconds exceeded."
                    % self.max_test_time}
        if interval > duration * 1000:
            return {'status': 'Error', 'message': "Interval must be <= duration."}
        if len(self.children) >= self.max_instances:
            return {'status': 'Error',
                    'message': "Too many concurrent instances running. "
                    "Try again later."}

        test_id = "".join(random.sample(ALPHABET, self.id_length))
        # Need one port for control, one for data (if the data stream is TCP).
        port = self.current_port
        self.current_port += 2
        return self._spawn_receiver(test_id, duration, interval, port, raw_data)

    def _clean_fork(self, output=None):
        pipe_r, pipe_w = os.pipe()
        pid = os.fork()
        if pid:
            self.children.append(pid)
            os.close(pipe_w)
            return pipe_r, False
        else:
            try:
                self.children = []
                self.toplevel = False
                os.chdir(self.working_dir)
                sys.stdin.close()
                # Move the pipe fd to position 3, close everything above it
                os.dup2(pipe_w, 3)
                os.closerange(4, MAXFD)
                pipe_w = 3

                if output is not None:
                    with open(os.path.join(self.working_dir, output), "w") as fp:
                        os.dup2(fp.fileno(), 1)
                        os.dup2(fp.fileno(), 2)
                else:
                    sys.stdout.close()
                    sys.stderr.close()
            except:
                try:
                    traceback.print_exc()
                finally:
                    os._exit(1)
            return pipe_w, True

    def _unlink(self, filename):
        if self.debug:
            return
        try:
            os.unlink(filename)
        except:
            pass

    def _spawn_receiver(self, test_id, duration, interval, port, raw_data):
        stdout = "%s.recv.out" % test_id
        datafile = self.datafile_pattern % test_id
        error = False
        pipe, child = self._clean_fork(stdout)
        if child:
            try:
                ret = self._run_receiver(
                    test_id, duration, interval, port, pipe, raw_data)
            except Exception as e:
                traceback.print_exc()
                ret = {'status': 'Error', 'message': str(e)}
                error = True
            try:
                # Write data to temp file, atomically rename to final file name
                with open('%s.tmp' % datafile, 'wt') as fp:
                    json.dump(ret, fp)
                    os.rename('%s.tmp' % datafile, datafile)
            finally:
                if not error:
                    self._unlink(stdout)
                    os._exit(0)
                os._exit(1)
        else:
            try:
                val = os.read(pipe, 1024).decode()
                if val == 'OK':
                    return {'status': 'OK', 'test_id': test_id, 'port': port}
                else:
                    return {'status': 'Error', 'message': val}
            finally:
                os.close(pipe)

    def _run_receiver(self, test_id, duration, interval, port, pipe, raw_data):
        logfile = '%s.log' % test_id
        txtlog = '%s.log.txt' % test_id
        outfile = '%s.dat' % test_id
        ret = {'status': 'OK'}

        # Run ITGRecv; does not terminate after the end of the test, so wait for
        # the test duration + a grace time of 5 seconds, then kill the process
        try:
            proc = subprocess.Popen(['ITGRecv',
                                     '-l', logfile,
                                     '-I',
                                     '-a', self.bind_address,
                                     '-Sp', str(port)])

        # Signal back to the parent whether or not we successfully spawned the
        # receiver.
        except OSError as e:
            os.write(pipe, str(e).encode())
        else:
            time.sleep(0.1)
            try:
                w = os.waitpid(proc.pid, os.WNOHANG)
            except OSError as e:
                w = (0, str(e))
            if w != (0, 0):
                ret = {'status': 'Error',
                       'message': 'ITGRecv exited immediately with code %s.'
                       % w[1]}
                os.write(pipe, ret['message'].encode())
            else:
                self.children.append(proc.pid)
                os.write(pipe, b'OK')

        os.close(pipe)

        # Use an alarm signal for the timeout; means we don't have to wait for the
        # entire test duration if the ITGRecv process terminates before then
        signal.alarm(duration + 5)
        try:
            retval = proc.wait()
            if retval > 0:
                raise Exception("ITGRecv non-zero exit")
        except AlarmException:
            proc.terminate()
        signal.alarm(0)

        # Call ITGDec on the log file output, read in the resulting data
        if not os.path.exists(logfile):
            raise Exception('No data file produced')

        try:
            subprocess.check_call(['ITGDec', logfile,
                                   '-c', str(interval), outfile,
                                   '-l', txtlog])
            with open(outfile, 'rt') as fp:
                ret['data'] = fp.read()

            if not ret['data']:
                # By not raising an exception, the stdout output file is still
                # cleared, but an error is reported to the caller. This is to
                # prevent output files from sticking around for, e.g., users
                # that request a test, but then cannot get packets through; the
                # output is not useful to diagnose that anyway.
                ret = {'status': 'Error', 'message': 'Empty data set.'}

            # Read start of text log file to get timestamp of first received
            # packet
            with open(txtlog, 'rt') as fp:
                if raw_data:
                    ret['raw'] = fp.read()
                    data = ret['raw'][:1024]
                else:
                    data = fp.read(1024)
                try:
                    idx_s = data.index('rxTime>') + len('rxTime>')
                    idx_e = data.index('Size>')
                    t, microsec = data[idx_s:idx_e].split(".")
                    h, m, s = t.split(":")
                    now = time.gmtime()
                    ret['utc_offset'] = calendar.timegm(
                        (now[0], now[1], now[2],
                         int(h), int(m), int(s),
                         now[6], now[7], now[8])) + int(microsec) / 10**6
                except Exception as e:
                    ret['utc_offset'] = None

            return ret
        finally:
            # Clean up temporary files and exit
            for f in logfile, txtlog, outfile:
                self._unlink(f)

    def _collect_garbage(self, *args):
        for p in self.children:
            try:
                if os.waitpid(p, os.WNOHANG) != (0, 0):
                    self.children.remove(p)
            except OSError:
                self.children.remove(p)
        # When no more children are alive, reset the current port
        if not self.children:
            self.current_port = self.start_port
        if not self.toplevel:
            return

        try:
            files = os.listdir(self.working_dir)
            for f in files:
                if not f.endswith('.json'):
                    continue
                p = os.path.join(self.working_dir, f)
                s = os.stat(p)
                td = datetime.now() - datetime.fromtimestamp(s.st_mtime)
                if td.seconds > 300:
                    self._unlink(p)
        except:
            traceback.print_exc()

    def _alarm(self, signum, frame):
        raise AlarmException()

    def _exit(self, signum, frame):
        for p in self.children:
            try:
                os.kill(p, signum)
            except OSError:
                pass

        # Do not call normal exit handlers when in a subprocess
        if not self.toplevel:
            os._exit(0)

        # SIGTERM is successful exit, others are not
        if signum == signal.SIGTERM:
            sys.exit(0)
        sys.exit(1)

    def __del__(self):
        if self.toplevel and not self.debug:
            shutil.rmtree(self.working_dir, ignore_errors=True)


def run():
    options, args = parser.parse_args()
    if not XML_DEFUSED and not options.INSECURE_XML:
        sys.stderr.write(
            "XML EXPANSION ATTACK VULNERABILITY DETECTED. ABORTING!\n"
            "Run with --insecure-xml to run anyway (will leave the server "
            "vulnerable!)\n")
        sys.exit(1)

    server = SimpleXMLRPCServer(
        (options.BIND_ADDRESS, options.BIND_PORT), allow_none=True)
    manager = DITGManager(
        bind_address=options.ITG_ADDRESS or options.BIND_ADDRESS,
        start_port=options.START_PORT,
        max_test_time=options.MAX_TEST_TIME,
        max_instances=options.MAX_INSTANCES,
        secret=options.SECRET,
        debug=options.DEBUG,)
    server.register_instance(manager)
    server.register_introspection_functions()
    server.serve_forever()


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
#
# test_utils.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     16 July 2015
# Copyright (c) 2015-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import unittest

from flent import util


class TestSmallUtilFunctions(unittest.TestCase):

    def test_uscore_to_camel(self):
        self.assertEqual(util.uscore_to_camel('test_name'), 'TestName')

    def test_classname(self):
        self.assertEqual(util.classname('test_class'), 'TestClass')

        self.assertEqual(util.classname(
            'test_class', 'Suffix'), 'TestClassSuffix')


test_suite = unittest.TestLoader().loadTestsFromTestCase(TestSmallUtilFunctions)



# -*- coding: utf-8 -*-
#
# util.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     16 October 2012
# Copyright (c) 2012-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import argparse
import fcntl
import fnmatch
import ipaddress
import json
import os
import re
import shlex
import shutil
import socket
import threading
import time
import subprocess

from copy import copy
from calendar import timegm
from datetime import datetime, timedelta
from math import log10, exp, sqrt

try:
    from datetime import UTC
except ImportError:
    UTC = None

from flent.loggers import get_logger

MULTIHOST_SEP = '//'
ENCODING = "UTF-8"
try:
    import locale
    ENCODING = locale.getpreferredencoding(False)
except:
    pass

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

logger = get_logger(__name__)


def uscore_to_camel(s):
    """Turn a underscore style string (org_table) into a CamelCase style string
    (OrgTable) for class names."""
    return ''.join(x.capitalize() for x in s.split("_"))


def classname(s, suffix=''):
    return uscore_to_camel(s) + suffix

def utcnow():
    if UTC is None:
        return datetime.utcnow()
    return datetime.now(UTC).replace(tzinfo=None)

def utcfromtimestamp(ts):
    if UTC is None:
        return datetime.utcfromtimestamp(ts)
    return datetime.fromtimestamp(ts, UTC).replace(tzinfo=None)

def format_date(dt, fmt="%Y-%m-%dT%H:%M:%S.%f", utc=False):
    if utc:
        return dt.strftime(fmt + "Z")
    # The datetime object is already UTC, so use gmtime rather than mktime to
    # get the timestamp from which to compute the UTC offset.
    ts = timegm(dt.timetuple()) + dt.microsecond / 1000000.0
    offset = datetime.fromtimestamp(ts) - utcfromtimestamp(ts)
    return (dt + offset).strftime(fmt)


def format_bytes(nbytes):
    if nbytes > 2**30:
        return (2**30, 'Gbytes')
    elif nbytes > 2**20:
        return (2**20, 'Mbytes')
    elif nbytes > 2**10:
        return (1**10, 'Kbytes')
    else:
        return (nbytes, 'bytes')


def parse_date(timestring, min_t=None, offset=None):
    try:
        # Try to parse the straight UTC time string (has a Z at the end)
        return datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S.%fZ"), None
    except ValueError:
        try:
            dt = datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S.%f")
        except ValueError:
            dt = datetime.strptime(timestring, "%Y-%m-%dT%H:%M:%S")

        if offset is None:
            # The timestamp is in local time, so we need to find the timezone
            # offset. Guess this by computing the nearest whole-hour offset for
            # from the minimum timestamp in the data series (passed as min_t)
            #
            # If not min_t is available, compute the offset from the system time
            # zone.
            #
            # Return the offset so it can be re-used the next time this function
            # is called.

            logger.debug("Non-UTC timestamp %s, computing timezone offset",
                         timestring)

            if min_t is not None:
                ts = orig_ts = timegm(dt.timetuple())
                hours = 0
                mult = 1 if ts < min_t else -1
                while abs(ts - min_t) > 1800:
                    hours -= 1 * mult
                    ts += 3600 * mult
                offset = timedelta(hours=hours)
                logger.debug("Guessed timezone offset of %d hours for "
                             "ts %f min_t %f diff %f",
                             hours, orig_ts, min_t, orig_ts-min_t)
            else:
                ts = time.mktime(dt.timetuple())
                offset = (datetime.fromtimestamp(ts) -
                          utcfromtimestamp(ts))
                logger.debug("Computed offset of %s from system timezone",
                             offset)

        return dt - offset, offset


def parse_int(val):
    try:
        try:
            return int(val)
        except ValueError:
            if val.strip().startswith("0x"):
                return int(val, 16)
            raise
    except (ValueError, AttributeError):
        raise ValueError("Invalid integer value: %s" % val)


def clean_path(path, allow_dirs=False):
    if allow_dirs:
        return re.sub("[^A-Za-z0-9_/-]", "_", path)
    else:
        return re.sub("[^A-Za-z0-9_-]", "_", path)


def long_substr(data, prefix_only=False):
    """Find the longest common substring between a list of strings.
    Optionally limit search to prefixes only.

    Brute force approach (i.e. not very efficient...).
    Based on https://stackoverflow.com/questions/2892931/longest-common-substring-from-more-than-two-strings-python"""  # noqa: E501
    substr = ''
    if len(data) > 1 and len(data[0]) > 0:
        if prefix_only:
            start_pos = [0]
        else:
            start_pos = range(len(data[0]))
        for i in start_pos:
            for j in range(len(data[0]) - i + 1):
                if j > len(substr) and all(data[0][i:i + j] in x for x in data):
                    substr = data[0][i:i + j]
    return substr


def diff_parts(strings, sep):
    """Return the unique parts of a set of strings by splitting on
    a separator and pruning parts that are identical for all strings"""

    parts = [s.split(sep) for s in strings]
    np = [p for p in zip(*parts) if len(set(p)) >= 1]

    return [sep.join(p) for p in zip(*np)]

class CachingDictionary(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filename = None
        self.lock = threading.Lock()

    def read_file(self, filename):
        try:
            with open(filename) as fp:
                fcntl.flock(fp, fcntl.LOCK_EX)
                if fp.read(1) == '':
                    # JSON parser chokes on an empty file, so turn this into a
                    # FileNotFoundError (which is handled below)
                    raise FileNotFoundError

                fp.seek(0)
                obj = json.load(fp)
                self.update(obj)
                logger.debug("Read cache '%s' from file '%s'", self, filename)
        except FileNotFoundError:
            if not self.write_file(filename):
                # File doesn't exist (or is empty), make sure we can write to it
                logger.warning("Couldn't write to cache file '%s'; not using.",
                               filename)
                return
            else:
                logger.debug("Created cache file at '%s'", filename)
        except (IOError, json.decoder.JSONDecodeError) as e:
            logger.warning("Error reading cache file '%s' (%s); not using.",
                           self.filename, e)
            return

        self.filename = filename

    def write_file(self, filename=None):
        if filename is None:
            filename = self.filename

        if filename is None:
            return

        try:
            with open(filename, 'w') as fp:
                fcntl.flock(fp, fcntl.LOCK_EX)
                json.dump(self, fp)
                return True
        except IOError as e:
            logger.debug("Writing cache file failed: '%s'", e)
            return False

    def __setitem__(self, key, value):
        with self.lock:
            super().__setitem__(key, value)
            self.write_file()

    def get_view(self, key):
        return CachingDictView(self, key)


class CachingDictView:

    def __init__(self, parent, key):
        if key not in parent:
            parent[key] = {}
        self.parent = parent
        self.key = key

    def __getitem__(self, key):
        return self.parent[self.key][key]

    def __setitem__(self, key, value):
        with self.parent.lock:
            self.parent[self.key][key] = value
            self.parent.write_file()

    def __delitem__(self, key):
        with self.parent.lock:
            del self.parent[self.key][key]
            self.parent.write_file()

    def __contains__(self, key):
        return key in self.parent[self.key]

    def __iter__(self):
        return iter(self.parent[self.key])

    def update(self, *args, **kwargs):
        with self.parent.lock:
            self.parent[self.key].update(*args, **kwargs)
            self.parent.write_file()

    def keys(self):
        return self.parent[self.key].keys()

    def values(self):
        return self.parent[self.key].values()

    def items(self):
        return self.parent[self.key].items()

global_cache = CachingDictionary()
get_cache = global_cache.get_view


class WhichCache:

    def __init__(self):
        self.cache = get_cache("which")

    @staticmethod
    def is_executable(filename):
        return os.path.isfile(filename) and os.access(filename, os.X_OK)

    @staticmethod
    def lookup_remote(executable, remote_host):
        logger.debug("running 'command -v' for binary '%s' on host '%s'",
                     executable, remote_host)
        try:
            output = subprocess.check_output(['ssh', remote_host,
                                              'command -v {}'.format(executable)],
                                             timeout=1)
            output = output.decode(ENCODING).strip()
            logger.debug("Got path '%s' for '%s' on '%s'",
                         output, executable, remote_host)
            return output
        except subprocess.CalledProcessError:
            pass

        return None

    def lookup_local(self, executable):
        pathname, filename = os.path.split(executable)
        if pathname:
            if not self.is_executable(executable):
                logger.debug("which: %s not executable", executable)
                return None

            logger.debug("which: %s is a full path and executable", executable)
            return executable

        filename = shutil.which(executable)
        if filename is not None:
            logger.debug("which: Found %s executable at %s",
                         executable, filename)
            return filename

        return None

    def __call__(self, executable, fail=None, remote_host=None):
        key = f"{executable},{remote_host or ''}"
        if key in self.cache:
            logger.debug("which: found '%s' in cache: %s", key, self.cache[key])
            return self.cache[key]

        if remote_host:
            binary = self.lookup_remote(executable, remote_host)
        else:
            binary = self.lookup_local(executable)

        if not binary and fail:
            raise fail("No %s binary found in PATH." % executable)

        if binary:
            self.cache[key] = binary
        return binary

# callers call 'util.which', so make that an instance of the cache
which = WhichCache()


def path_components(path):
    folders = []
    while path and path != "/":
        path, folder = os.path.split(path)

        if folder:
            folders.insert(0, folder)
    if path == "/":
        folders.insert(0, path)
    return folders

def normalise_host(hostname, bracket_v6=False):
    if hostname and MULTIHOST_SEP in hostname:
        hostname = hostname.split(MULTIHOST_SEP, 1)[0]
    if bracket_v6:
        try:
            ipaddress.IPv6Address(hostname)
            hostname = "[{}]".format(hostname)
        except ValueError:
            pass
    return hostname

def lookup_host(hostname, version=None):
    hostname = normalise_host(hostname)
    logger.debug("Looking up hostname '%s'.", hostname)
    if version == 4:
        version = socket.AF_INET
    elif version == 6:
        version = socket.AF_INET6
    else:
        version = socket.AF_UNSPEC
    try:
        hostnames = socket.getaddrinfo(hostname, None, version,
                                       socket.SOCK_STREAM)
    except Exception:
        hostnames = None

    if not hostnames:
        raise RuntimeError("Found no hostnames on lookup of %s" % hostname)

    return hostnames[0]


class DefaultConfigParser(configparser.ConfigParser):

    class _NoDefault(object):
        pass

    def get(self, section, option, default=_NoDefault):
        try:
            return configparser.ConfigParser.get(self, section, option)
        except configparser.NoOptionError:
            if default == self._NoDefault:
                raise
            else:
                return default

    def getint(self, section, option, default=_NoDefault):
        try:
            return configparser.ConfigParser.getint(self, section, option)
        except configparser.NoOptionError:
            if default == self._NoDefault:
                raise
            else:
                return default

    def getfloat(self, section, option, default=_NoDefault):
        try:
            return configparser.ConfigParser.getfloat(self, section, option)
        except configparser.NoOptionError:
            if default == self._NoDefault:
                raise
            else:
                return default

    def getboolean(self, section, option, default=_NoDefault):
        try:
            return configparser.ConfigParser.getboolean(self, section, option)
        except configparser.NoOptionError:
            if default == self._NoDefault:
                raise
            else:
                return default


class Glob(object):
    """Object for storing glob patterns in matches"""

    def __init__(self, pattern, exclude=None):
        if exclude is None:
            self.exclude = []
        else:
            self.exclude = exclude
        self.pattern = pattern

    def __repr__(self):
        return "<Glob: %s (excl: %s)>" % (self.pattern, ",".join(self.exclude))

    def filter(self, values, exclude, args=None):
        if args is not None:
            pattern = self.pattern.format(**args)
        else:
            pattern = self.pattern

        # Exclude * from matching :, make ** match everything
        re_pat = fnmatch.translate(pattern.replace("**", "___PLACEHOLDER___"))
        re_pat = re_pat.replace(".*", "[^:]*")
        re_pat = re_pat.replace("___PLACEHOLDER___", ".*")
        regex = re.compile(re_pat)

        exclude += self.exclude
        return (x for x in values if regex.match(x) and
                not any((fnmatch.fnmatch(x, e) for e in exclude)))

    def __iter__(self):
        return iter((self,))  # allow list(g) to return [g]

    @classmethod
    def filter_dict(cls, d, args=None):
        # Expand glob patterns in parameters. Go through all items in the
        # dictionary looking for subkeys that is a Glob instance or a list
        # that has a Glob instance in it.
        for k, v in list(d.items()):
            for g_k in list(v.keys()):
                try:
                    v[g_k] = cls.expand_list(
                        v[g_k], list(d.keys()), [k], args=args)
                except TypeError:
                    continue
        return d

    @classmethod
    def expand_list(cls, l, values, exclude=None, args=None):
        l = list(l)  # copy list, turns lone Glob objects into [obj]
        new_l = []
        if exclude is None:
            exclude = []
        # Expand glob patterns in list. Go through all items in the
        # list  looking for Glob instances and expanding them.
        for pattern in l:
            if isinstance(pattern, cls):
                new_l.extend(pattern.filter(values, exclude, args))
            else:
                new_l.append(pattern)
        return new_l


def mos_score(T, loss):
    """Calculate a MOS score based on a one-way delay and a packet loss rate.
    Based on ITU G.107 06/2015.

    Verified against the online reference implementation at
    https://www.itu.int/ITU-T/studygroups/com12/emodel/calculnb.php

    This version assumes the default values are used for all parameters other
    than delay and loss.

    @T: Mean one-way delay
    @loss: Packet loss rate between 0 and 1.

    """

    # All variable names are from G.107.

    # Parameters
    Ta = T
    Tr = 2 * T
    Ppl = loss * 100  # in percent

    # Defaults
    mT = 100  # Table 1

    # From Table 3:
    WEPL = 110
    TELR = 65
    RLR = 2
    SLR = 8

    # Constants calculated from the Table 3 defaults:
    No = -61.17921438624169  # (7-3)
    Ro = 15 - (1.5 * (SLR + No))  # (7-2)
    Is = 1.4135680813438616  # (7-8)

    Rle = 10.5 * (WEPL + 7) * pow(Tr + 1, -0.25)  # (7-26)
    if Ta == 0:
        X = 0
    else:
        X = log10(Ta / mT) / log10(2)  # (7-28)

    if Ta <= 100:
        Idd = 0
    else:
        Idd = 25 * ((1 + X**6)**(1 / 6) - 3 *
                    (1 + (X / 3)**6)**(1 / 6) + 2)  # (7-27)

    Idle = (Ro - Rle) / 2 + sqrt((Ro - Rle)**2 / 4 + 169)  # (7-25)

    TERV = TELR - 40 * log10((1 + T / 10) / (1 + T / 150)) + \
        6 * exp(-0.3 * T**2)  # (7-22)
    Roe = -1.5 * (No - RLR)  # (7-20)
    Re = 80 + 2.5 * (TERV - 14)  # (7-21)

    if T < 1:
        Idte = 0
    else:
        Idte = ((Roe - Re) / 2 + sqrt((Roe - Re)**2 / 4 + 100) - 1) * \
            (1 - exp(-T))  # (7-19)

    Id = Idte + Idle + Idd  # (7-18)

    Ieeff = 95 * (Ppl / (Ppl + 4.3))  # (7-29) with BurstR = 1, Bpl = 4.3

    R = Ro - Is - Id - Ieeff

    if R < 0:
        MOS = 1
    elif R > 100:
        MOS = 4.5
    else:
        MOS = 1 + 0.035 * R + R * (R - 60) * (100 - R) * 7 * 10**-6  # (B-4)

    return MOS


# Argparse stuff

class FuncAction(argparse.Action):

    def __init__(self, option_strings, dest, help=None):
        super(FuncAction, self).__init__(option_strings,
                                         dest,
                                         nargs=0,
                                         required=False,
                                         help=help)


class Update(argparse.Action):

    def __init__(self, *args, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = {}
        super(Update, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        if not hasattr(namespace, self.dest):
            setattr(namespace, self.dest, self.default)
        getattr(namespace, self.dest).update(values)


def _copy_items(items):
    if items is None:
        return []
    # The copy module is used only in the 'append' and 'append_const'
    # actions, and it is needed only when the default value isn't a list.
    # Delay its import for speeding up the common case.
    if type(items) is list:
        return items[:]
    return copy(items)

def append_host(items, new_host):
    host, suffix, idx = (None, None, None)
    for k, v in enumerate(items):
        if MULTIHOST_SEP in v:
            h, s = v.split(MULTIHOST_SEP, 1)
            s = int(s)
        else:
            h, s = (v, 1)
        if h == new_host:
            host, suffix, idx = h, s, k
    if host:
        new_host = MULTIHOST_SEP.join((new_host, str(suffix+1)))
        items[idx] = MULTIHOST_SEP.join((host, str(suffix)))
    items.append(new_host)

class AddHost(argparse._AppendAction):

    def __init__(self, *args, **kwargs):
        if 'default' not in kwargs:
            kwargs['default'] = []
        super(AddHost, self).__init__(*args, **kwargs)

    def __call__(self, parser, namespace, new_host, option_string=None):
        items = getattr(namespace, self.dest, None)
        items = _copy_items(items)
        append_host(items, new_host)
        setattr(namespace, self.dest, items)

def float_pair(value):
    try:
        if "," not in value:
            return (None, float(value))
        a, b = [s.strip() for s in value.split(",", 1)]
        return (float(a) if a else None,
                float(b) if b else None)
    except ValueError:
        raise argparse.ArgumentTypeError("Invalid pair value: %s" % value)


def float_pair_noomit(value):
    val = float_pair(value)
    if None in val:
        raise argparse.ArgumentTypeError("Invalid pair value: %s" % value)
    return val


def keyval(value):
    ret = {}
    for p in token_split(value, ";"):
        p = p.strip()
        if not p:
            continue
        try:
            k, v = token_split(p, '=')
            ret.update({k.strip(): v.strip()})
        except ValueError:
            raise argparse.ArgumentTypeError(
                "Invalid value '%s' (missing =)" % p)
    return ret

def noop(x):
    return x

def todict(k, v):
    return dict(k=v)

def rangedict(key, value):
    ret = {}
    for k in key.split(","):
        if '-' in k:
            s, e = (int(i) for i in k.split("-", 1))
            if not s < e:
                raise ValueError
            for i in range(s, e+1):
                ret[i] = value
        elif k == '*':
            ret['*'] = value
        else:
            ret[int(k)] = value
    return ret

def keyval_pair_transformer(pairfunc=todict, errmsg="Parse error"):
    def typefunc(value):
        ret = {}
        try:
            for k, v in keyval(value).items():
                ret.update(pairfunc(k, v))
            return ret
        except ValueError:
            raise argparse.ArgumentTypeError(errmsg)
    return typefunc

def keyval_transformer(keyfunc=noop, valfunc=noop, errmsg="Parse error"):
    return keyval_pair_transformer(pairfunc=lambda k, v: {keyfunc(k): valfunc(v)},
                                   errmsg=errmsg)

keyval_int = keyval_pair_transformer(pairfunc=rangedict,
                                     errmsg="Keys must be integers.")


def comma_list(value):
    try:
        return [v.strip() for v in token_split(value)]
    except ValueError:
        raise argparse.ArgumentTypeError("Unable to split into list.")


def token_split(value, split_tokens=',;'):
    """Split VALUE on the tokens given in SPLIT_TOKENS, while
    avoiding splitting of quoted strings"""

    lex = shlex.shlex(value, posix=True)
    lex.whitespace_split = True
    lex.whitespace = split_tokens
    return list(lex)


class ArgParam(object):
    """A class that takes an argparser and sets object properties from
    the argparser-defined parameters."""

    params = None

    def __init__(self, **kwargs):
        if self.params:
            for a in self.params._actions:
                dest = a.dest.lower()
                if dest in kwargs:
                    setattr(self, dest, copy(kwargs[dest]))
                elif a.dest in kwargs:
                    setattr(self, dest, copy(kwargs[a.dest]))
                else:
                    setattr(self, dest, copy(a.default))


class ArgParser(argparse.ArgumentParser):

    def get_type(self, dest):
        for action in self._actions:
            if action.dest == dest:
                # Workaround because StoreConst actions don't store the action
                # type
                if isinstance(action, argparse._StoreConstAction):
                    return type(action.const)
                return action.type
        return None

    def get_choices(self, dest):
        for action in self._actions:
            if action.dest == dest and action.choices:
                return action.choices

        return None

    def is_list(self, dest):
        for action in self._actions:
            if action.dest == dest:
                return isinstance(action, argparse._AppendAction)
        return False

    def __contains__(self, dest):
        for action in self._actions:
            if action.dest == dest:
                return True
        return False



# -*- coding: utf-8 -*-
#
# __init__.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:      6 December 2012
# Copyright (c) 2012-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import locale
import os
import signal
import sys


# Convert SIGTERM into SIGINT to apply the same shutdown logic.
def handle_sigterm(sig, frame):
    os.kill(os.getpid(), signal.SIGINT)


def run_flent(gui=False):
    if sys.version_info[:2] < (3, 6):
        sys.stderr.write("Sorry, Flent requires v3.6 or later of Python.\n")
        sys.exit(1)
    try:
        try:
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error:
            pass
        from flent import batch
        from flent.settings import load
        from flent.loggers import setup_console, get_logger

        setup_console()
        logger = get_logger(__name__)
        logger.debug("Flent executed as %s in PID %d", sys.argv, os.getpid())

        try:
            signal.signal(signal.SIGTERM, handle_sigterm)
            settings = load(sys.argv[1:])
            if gui or settings.GUI:
                from flent.gui import run_gui
                return run_gui(settings)
            else:
                b = batch.new(settings)
                b.run()

        except RuntimeError as e:
            logger.exception(str(e))

    except KeyboardInterrupt:
        try:
            b.kill()
        except NameError:
            pass

        # Proper behaviour on SIGINT is re-killing self with SIGINT to properly
        # signal to surrounding shell what happened.
        # Ref: http://mywiki.wooledge.org/SignalTrap
        try:
            signal.signal(signal.SIGINT, signal.SIG_DFL)
            os.kill(os.getpid(), signal.SIGINT)
        except:
            return 1  # Just in case...
    finally:
        try:
            signal.signal(signal.SIGTERM, signal.SIG_DFL)
        except:
            pass
    return 0


def run_flent_gui():
    return run_flent(gui=True)


__all__ = ['run_flent', 'run_flent_gui']



# -*- coding: utf-8 -*-
#
# __init__.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     16 July 2015
# Copyright (c) 2015-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import unittest
from . import test_util
from . import test_formatters
from . import test_metadata
from . import test_parsers
from . import test_plotters
from . import test_tests
from . import test_gui


test_suite = unittest.TestSuite([test_util.test_suite,
                                 test_formatters.test_suite,
                                 test_metadata.test_suite,
                                 test_parsers.test_suite,
                                 test_plotters.test_suite,
                                 test_tests.test_suite,
                                 test_gui.test_suite,
                                 ])

all_tests = unittest.TestSuite([test_suite, test_plotters.plot_suite])

def load_tests(loader, standard_tests, pattern):
    suite = os.getenv("TEST_SUITE", None)
    if suite == "all_tests":
        return all_tests
    return test_suite

if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
#
# tranformers.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     24 October 2012
# Copyright (c) 2012-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

try:
    import numpy as np
except ImportError:
    np = None


def transform_results(results, func):
    """Transform a list of (timestamp,value) pairs by applying a function to the
    value."""
    try:
        for i in range(len(results)):
            results[i][1] = func(results[i][1])
    except TypeError:
        # Transformers can also be applied to metadata items, in which case
        # calling len() on results will result in a TypeError. So just try to
        # call the function on the value itself.
        return func(results)
    return results


def rr_to_ms(results):
    """Transforms a transactions/second netperf RR measurement into ping times
    in milliseconds."""

    if hasattr(results, "shape"):
        return 1000 / results

    def safe_divide(x):
        if x == 0:
            return None
        return 1000.0 / x
    return transform_results(results, safe_divide)


def s_to_ms(results):
    if hasattr(results, "shape"):
        return results * 1000.0
    return transform_results(results, lambda x: x * 1000.0)


def bits_to_mbits(results):
    if hasattr(results, "shape"):
        return results / 1000000.0
    return transform_results(results, lambda x: x / 1000000.0)


def kbits_to_mbits(results):
    if hasattr(results, "shape"):
        return results / 1000.0
    return transform_results(results, lambda x: x / 1000.0)


def cumulative_to_events(results):
    """Transform cumulative counter values into the increasing events."""
    if hasattr(results, "shape") and np is not None:
        # Need output array same length as input array
        arr = np.zeros(len(results), dtype=float)
        arr[1:] = np.diff(results)
        return arr

    try:
        current = results[0][1]
        res = []
        for t, v in results:
            res.append((t, v - current))
            current = v
        return res
    except (TypeError, IndexError):
        return results


def identity(results):
    return results



# -*- coding: utf-8 -*-
#
# __main__.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     21 May 2015
# Copyright (c) 2015-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from multiprocessing import freeze_support

from flent import run_flent

if __name__ == "__main__":
    freeze_support()
    sys.exit(run_flent())



# -*- coding: utf-8 -*-
#
# batch.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     11 April 2014
# Copyright (c) 2014-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
import os
import pprint
import random
import re
import signal
import subprocess
import sys
import time
import uuid

from datetime import timedelta
from fnmatch import fnmatch
from collections import OrderedDict

try:
    from configparser import RawConfigParser
except ImportError:
    from ConfigParser import RawConfigParser

from flent import aggregators, formatters, resultset, loggers
from flent.metadata import record_metadata, record_postrun_metadata
from flent.util import clean_path, format_date, token_split, utcnow
from flent.settings import parser as SETTINGS_PARSER
from flent.settings import _LOG_DEFER

# Python2/3 compatibility
try:
    basestring
except NameError:
    basestring = str

logger = loggers.get_logger(__name__)


def new(settings):
    return BatchRunner(settings)


class BatchRunner(object):

    _INTERP_REGEX = re.compile(r"(^|[^$])(\$\{([^}]+)\})")
    _MAX_INTERP = 1000

    def __init__(self, settings):
        self.args = OrderedDict()
        self.batches = OrderedDict()
        self.commands = OrderedDict()
        self.settings = settings
        self.killed = False
        self.children = []
        self.logfile = self.logfile_debug = None
        self.tests_run = 0

        for f in settings.BATCH_FILES:
            self.read(f)

    def read(self, filename):
        parser = RawConfigParser(dict_type=OrderedDict)
        read = parser.read(filename)
        if read != [filename]:
            raise RuntimeError("Unable to read batch file: %s." % filename)

        for s in parser.sections():
            typ, nam = s.split("::")
            if typ.lower() == 'arg':
                self.args[nam.lower()] = OrderedDict(parser.items(s))
            elif typ.lower() == 'batch':
                self.batches[nam.lower()] = OrderedDict(parser.items(s))
            elif typ.lower() == 'command':
                self.commands[nam.lower()] = OrderedDict(parser.items(s))
            else:
                raise RuntimeError("Unknown section type: '%s'." % typ)

        self.expand_groups()

    def expand_groups(self):
        for obj in self.args, self.batches, self.commands:
            for name, vals in obj.items():
                # Expand inheritance
                if 'inherits' in vals:
                    for inh in [x.strip() for x in reversed(vals['inherits'].split(','))]:
                        if not inh in obj:
                            raise RuntimeError(
                                "%s inherits from non-existent parent %s."
                                % (name, inh))
                        obj[name] = self.inherit(obj[inh], obj[name])

                # Parse boolean options
                for k, v in obj[name].items():
                    if isinstance(v, basestring) and v.lower() in ('yes', 'true', 'on'):  # noqa: E501
                        obj[name][k] = True
                    elif isinstance(v, basestring) and v.lower() in ('no', 'false', 'off'):  # noqa: E501
                        obj[name][k] = False

    def inherit(self, parent, child):
        new = parent.copy()

        # Make sure children are not declared abstract.
        if 'abstract' in new:
            del new['abstract']

        new.update(child)
        if 'inherits' in parent:
            new['inherits'] = "%s, %s" % (parent['inherits'], child['inherits'])

        return new

    def get_ivar(self, name, ivars, settings):
        if name in ivars:
            return str(ivars[name])
        elif hasattr(settings, name.upper()):
            return str(getattr(settings, name.upper()))
        else:
            return "$${%s}" % name

    def interpolate(self, string, ivars, settings=None):
        """Perform recursive expansion of ${vars}.

        Works by looking for a string matching the expansion syntax and
        replacing that with the value of the ivars dict corresponding to the key
        inside {}. If no key matching key is found, the expansion is escaped (by
        duplicating the $), to make sure the expansion ends. Cyclic expansions
        are protected against by capping the number of iterations."""

        if not isinstance(string, basestring):
            return string

        if settings is None:
            settings = self.settings

        ret = string
        m = self._INTERP_REGEX.search(ret)
        i = 0
        while m is not None:
            k = m.group(3)
            ret = ret.replace(m.group(2), self.get_ivar(k, ivars, settings))
            m = self._INTERP_REGEX.search(ret)
            i += 1
            if i > self._MAX_INTERP:
                raise RuntimeError(
                    "Cyclic interpolation (more than %d expansions performed)."
                    % self._MAX_INTERP)

        return ret.replace("$$", "$")

    def apply_args(self, values, args={}, settings=None):
        new = OrderedDict(values)
        new.update(args)
        for k, v in new.items():
            new[k] = self.interpolate(v, new, settings)

        return new

    def commands_for(self, batch, settings=None):
        if 'commands' not in batch:
            return []
        cmdnames = [i.strip() for i in token_split(batch['commands'])]
        commands = OrderedDict()

        while cmdnames:
            c = cmdnames.pop(0)
            if not c or c in commands:
                continue
            if c not in self.commands:
                raise RuntimeError("Can't find command '%s' when expanding "
                                   "batch command." % c)
            cmd = self.apply_args(self.commands[c], batch, settings)

            # Don't include disabled commands
            if cmd.get('disabled', False) or not cmd.get('enabled', True):
                continue

            # Commands can specify extra commands to run; expand those, use the
            # dictionary to prevent duplicates
            extra = [i.strip() for i in token_split(cmd.get('extra_commands', ''))
                     if i.strip()]
            for e in reversed(extra):
                cmdnames.insert(0, e)
            commands[c] = cmd

        return commands.values()

    def run_command(self, command):
        cmd = command['exec'].strip()
        if self.settings.BATCH_VERBOSE:
            ess = command.get('essential', False)
            if self.settings.BATCH_DRY:
                logger.info("  Would run '%s' (%s)", cmd,
                            'essential' if ess else 'non-essential')
            else:
                logger.info("  Running '%s', (%s)", cmd,
                            'essential' if ess else 'non-essential')

        if self.settings.BATCH_DRY:
            return
        if command['type'] in ('pre', 'post'):
            try:
                res = subprocess.check_output(cmd, universal_newlines=True,
                                              shell=True,
                                              stderr=subprocess.STDOUT)
                logger.debug("Command '%s' executed successfully.", cmd,
                             extra={'output': res})
            except subprocess.CalledProcessError as e:
                if command.get('essential', False):
                    raise RuntimeError("Essential command '%s' failed. "
                                       "Return code: %s.\nOutput:\n %s."
                                       % (cmd, e.returncode,
                                          "\n ".join(e.output.splitlines())))
                else:
                    logger.warning("Command '%s' failed (%d).",
                                   cmd,
                                   e.returncode,
                                   extra={'output': e.output})
        elif command['type'] in ('monitor',):
            proc = subprocess.Popen(cmd, universal_newlines=True, shell=True,
                                    stderr=subprocess.STDOUT)
            self.children.append((proc, command.get('kill', False)))

    def kill_children(self, force=False):
        if not self.children:
            return

        logger.debug("Killing child processes")
        for proc, kill in self.children:
            if kill or force:
                proc.terminate()
            else:
                proc.wait()
        self.children = []

    def run_commands(self, commands, ctype, essential_only=False):
        for c in commands:
            if c['type'] == ctype and (not essential_only or
                                       c.get('essential', False)):
                try:
                    self.run_command(c)
                except:
                    # When running in essential only mode, suppress errors to
                    # make sure subsequent commands don't fail
                    if not essential_only:
                        raise

    def gen_filename(self, settings, batch, argset, rep):
        p = ["batch", settings.BATCH_NAME]
        if self.settings.BATCH_TIMESTAMP:
            p.append(batch['batch_time'])
        p.append(batch.get('filename_extra', "%s-%s" % (argset, rep)))
        return clean_path("-".join(p))

    def expand_argsets(self, batch, argsets, batch_time, batch_name,
                       print_status=True, no_shuffle=False):

        sets = itertools.product(*argsets)
        if self.settings.BATCH_SHUFFLE and not no_shuffle:
            sets = list(sets)
            random.shuffle(sets)

        for argset in sets:
            rep = argset[-1]
            argset = argset[:-1]
            settings = self.settings.copy()
            if print_status:
                logger.info(" args:%s rep:%02d%s", ",".join(argset), rep,
                            " (dry run)" if settings.BATCH_DRY else "")
            settings.FORMAT = 'null'
            settings.BATCH_NAME = batch_name
            settings.BATCH_TIME = batch_time
            settings.TIME = utcnow()

            expand_vars = {'repetition': "%02d" % rep,
                           'batch_time': format_date(settings.BATCH_TIME,
                                                     fmt="%Y-%m-%dT%H%M%S")}

            for arg in argset:
                if arg not in self.args:
                    raise RuntimeError("Invalid arg: '%s'." % arg)
                expand_vars.update(self.args[arg])

            b = self.apply_args(batch, expand_vars, settings)

            if 'test_name' not in b:
                raise RuntimeError("Missing test name.")

            settings.load_rcvalues(b.items(), override=True)
            settings.NAME = b['test_name']
            settings.load_test(informational=True)
            settings.DATA_FILENAME = self.gen_filename(settings, b, argset, rep)

            yield b, settings

    def get_argsets(self, batch):
        argsets = []

        for k in batch.keys():
            if k.startswith("for_"):
                argset = []
                for a in token_split(batch[k]):
                    a = a.strip().lower()
                    matches = [arg for arg in self.args if fnmatch(arg, a)]
                    if not matches:
                        raise RuntimeError("No matches for arg: '%s'." % a)
                    argset.extend(matches)
                argsets.append(argset)

        reps = range(1, int(batch.get('repetitions', 1)) + 1)
        argsets.append(reps)

        return argsets

    def get_batch_runtime(self, batch_name):
        if batch_name not in self.batches:
            raise RuntimeError("Can't find batch '%s' to run." % batch_name)
        batch = self.batches[batch_name].copy()
        batch.update(self.settings.BATCH_OVERRIDE)

        if batch.get('abstract', False) or batch.get('disabled', False):
            return (0, 0)

        total_time = 0
        n = 0
        argsets = self.get_argsets(batch)

        for _, s in self.expand_argsets(batch, argsets, self.settings.TIME,
                                        batch_name, print_status=False,
                                        no_shuffle=True):
            total_time += s.TOTAL_LENGTH + int(batch.get('pause', 0))
            n += 1

        return total_time, n

    def run_batch(self, batch_name):
        if batch_name not in self.batches:
            raise RuntimeError("Can't find batch '%s' to run." % batch_name)
        batch = self.batches[batch_name].copy()
        batch.update(self.settings.BATCH_OVERRIDE)

        # A batch declared 'abstract' is not runnable
        if batch.get('abstract', False):
            logger.info(" Batch marked as abstract. Not running.")
            return False
        elif batch.get('disabled', False) or not batch.get('enabled', True):
            logger.info(" Batch disabled.")
            return False

        argsets = self.get_argsets(batch)

        pause = int(batch.get('pause', 0))

        batch_time = None
        if self.settings.BATCH_RESUME is not None and \
           os.path.isdir(self.settings.BATCH_RESUME):
            # We're resuming a batch run. Try to find a data file we can get the
            # original batch run time from.
            for dirpath, dirnames, filenames in os.walk(
                    self.settings.BATCH_RESUME):
                datafiles = [f for f in filenames if f.endswith(resultset.SUFFIX)]
                if datafiles:
                    f = datafiles[0]
                    r = resultset.load(os.path.join(dirpath, f))
                    batch_time = r.meta("BATCH_TIME")
                    try:
                        self.settings.BATCH_UUID = r.meta("BATCH_UUID")
                        logger.info(" Using previous UUID %s.\n",
                                    self.settings.BATCH_UUID)
                    except KeyError:
                        pass
                    break
            if batch_time is None:
                raise RuntimeError("No data files found in resume directory %s."
                                   % self.settings.BATCH_RESUME)
        elif self.settings.BATCH_RESUME:
            raise RuntimeError("Batch resume directory %s doesn't exist!\n"
                               % self.settings.BATCH_RESUME)
        else:
            batch_time = self.settings.TIME

        filenames_seen = set()

        for b, settings in self.expand_argsets(batch, argsets,
                                               batch_time, batch_name):

            if 'output_path' in b:
                output_path = clean_path(b['output_path'], allow_dirs=True)
            else:
                output_path = settings.DATA_DIR

            if settings.BATCH_RESUME is not None:
                if os.path.commonprefix(
                    [os.path.abspath(output_path),
                     os.path.abspath(settings.BATCH_RESUME)]) \
                        != os.path.abspath(settings.BATCH_RESUME):
                    raise RuntimeError("Batch-specified output path is not a "
                                       "subdirectory of resume path. Bailing.")
                if os.path.exists(os.path.join(output_path, "%s%s"
                                               % (settings.DATA_FILENAME,
                                                  resultset.SUFFIX))):
                    logger.info("  Previous result exists, skipping.")
                    continue

            if settings.BATCH_DRY and settings.BATCH_VERBOSE:
                logger.info("  Would output to: %s.\n", output_path)
            elif not settings.BATCH_DRY and not os.path.exists(output_path):
                try:
                    os.makedirs(output_path)
                except OSError as e:
                    raise RuntimeError("Unable to create output path '%s': %s."
                                       % (output_path, e))

            commands = self.commands_for(b, settings)
            if not settings.BATCH_DRY:
                self.logfile = loggers.setup_logfile(
                    os.path.join(output_path,
                                 "%s.log" % settings.DATA_FILENAME),
                    level=loggers.INFO, replay=False)
                if b.get('debug_log', False):
                    self.logfile_debug = loggers.setup_logfile(
                        os.path.join(output_path,
                                     "%s.debug.log" % settings.DATA_FILENAME),
                        level=loggers.DEBUG,
                        maxlevel=loggers.DEBUG,
                        replay=False)

            if settings.DATA_FILENAME in filenames_seen:
                logger.warning("Filename already seen in this run: %s",
                               settings.DATA_FILENAME)
            filenames_seen.add(settings.DATA_FILENAME)

            self.run_commands(commands, 'pre')
            self.run_commands(commands, 'monitor')
            try:
                if settings.BATCH_VERBOSE:
                    if settings.BATCH_DRY:
                        logger.info("  Would run test '%s'.", settings.NAME)
                    else:
                        logger.info("  Running test '%s'.", settings.NAME)
                    logger.info("   data_filename=%s", settings.DATA_FILENAME)
                    for k in sorted(b.keys()):
                        if k.upper() in SETTINGS_PARSER:
                            logger.info("   %s=%s", k, b[k])

                if settings.BATCH_DRY:
                    self.tests_run += 1
                else:
                    # Load test again with informational=False to enable host
                    # lookups and other actions that may fail
                    settings.load_test(informational=False)
                    self.run_test(settings, output_path)
            except KeyboardInterrupt:
                self.run_commands(commands, 'post', essential_only=True)
                raise
            except Exception as e:
                self.run_commands(commands, 'post', essential_only=True)
                logger.exception("  Error running test: %s", str(e))
            else:
                try:
                    self.run_commands(commands, 'post')
                except Exception as e:
                    self.run_commands(commands, 'post', essential_only=True)
                    logger.exception("  Error running post-commands: %s", str(e))
            finally:
                self.kill_children()
                if self.logfile:
                    loggers.remove_log_handler(self.logfile)
                    self.logfile.close()
                    self.logfile = None
                if self.logfile_debug:
                    loggers.remove_log_handler(self.logfile_debug)
                    self.logfile_debug.close()
                    self.logfile_debug = None

            if settings.BATCH_DRY and settings.BATCH_VERBOSE:
                logger.info("  Would sleep for %d seconds.", pause)
            elif not settings.BATCH_DRY:
                time.sleep(pause)

    def run_test(self, settings, output_path, print_datafile_loc=False):
        settings = settings.copy()
        settings.load_test()
        res = resultset.new(settings)

        if settings.LOG_FILE is _LOG_DEFER:
            settings.LOG_FILE = res.dump_filename.replace(res.SUFFIX, ".log")
            loggers.setup_logfile(settings.LOG_FILE)

        record_metadata(res, settings.EXTENDED_METADATA,
                        settings.REMOTE_METADATA)

        if not settings.HOSTS:
            raise RuntimeError("Must specify host (-H option).")

        logger.info("Starting %s test. Expected run time: %d seconds.",
                    settings.NAME, settings.TOTAL_LENGTH)

        agg = aggregators.new(settings)
        res = agg.postprocess(agg.aggregate(res))

        record_postrun_metadata(res, settings.EXTENDED_METADATA,
                                settings.REMOTE_METADATA)
        res.dump_dir(output_path)
        logger.log(loggers.INFO if print_datafile_loc else loggers.DEBUG,
                   "Data file written to %s", res.dump_filename)

        formatter = formatters.new(settings)
        formatter.format([res])

        self.tests_run += 1

    def load_input(self, settings):
        settings = settings.copy()
        results = []
        test_name = None
        for i, filename in enumerate(settings.INPUT):
            r = resultset.load(filename, settings.ABSOLUTE_TIME)
            if test_name is not None and test_name != r.meta("NAME") and \
               not settings.GUI:
                logger.warning("Result sets are not from the same "
                               "test (found %s/%s).",
                               test_name,
                               r.meta("NAME"))
            test_name = r.meta("NAME")
            if results and settings.CONCATENATE:
                results[0].concatenate(r)
            else:
                if len(settings.OVERRIDE_LABELS) > i:
                    r.set_label(settings.OVERRIDE_LABELS[i])
                results.append(r)

        settings.update(results[0].meta())
        settings.load_test(informational=True)

        # Look for missing data series, and if they are computed from other
        # values, try to compute them.
        for res in results:
            settings.compute_missing_results(res)

        formatter = formatters.new(settings)
        formatter.format(results)

    def fork_and_run(self, queue):
        pid = os.fork()
        if pid:
            return pid
        else:
            loggers.set_queue_handler(queue)
            signal.signal(signal.SIGTERM, self.kill)
            try:
                self.run()
            except Exception as e:
                logger.exception(str(e))
            queue.close()
            queue.join_thread()
            os._exit(0)

    def run(self):
        if self.settings.INPUT:
            return self.load_input(self.settings)
        elif self.settings.BATCH_NAMES:
            start_time = self.settings.TIME
            self.settings.BATCH_UUID = str(uuid.uuid4())
            self.settings.LOG_FILE = None  # batch run will generate logs
            logger.info("Started batch run %s at %s.",
                        self.settings.BATCH_UUID,
                        format_date(start_time, fmt="%Y-%m-%d %H:%M:%S"))

            if len(self.settings.BATCH_NAMES) == 1 and \
               self.settings.BATCH_NAMES[0] == 'ALL':
                logger.info("Running all batches.")
                batches = self.batches.keys()
            else:
                batches = self.settings.BATCH_NAMES
            runtimes = [self.get_batch_runtime(b) for b in batches]
            total_time, total_n = map(sum, zip(*runtimes))
            if total_time > 0:
                logger.info("Estimated total runtime: %s (%d tests)",
                            timedelta(seconds=total_time),
                            total_n)
            for b in batches:
                try:
                    logger.info("Running batch '%s'.", b)
                    self.run_batch(b)
                except RuntimeError:
                    raise
                except Exception as e:
                    if self.settings.DEBUG_ERROR:
                        raise
                    raise RuntimeError("Error while running batch '%s': %r."
                                       % (b, e))
            end_time = utcnow()
            logger.info("Ended batch sequence at %s. %s %d tests in %s.",
                        format_date(end_time, fmt="%Y-%m-%d %H:%M:%S"),
                        "Ran" if not self.settings.BATCH_DRY else 'Would have run',  # noqa: E501
                        self.tests_run,
                        (end_time - start_time))
            return True
        else:
            return self.run_test(self.settings, self.settings.DATA_DIR, True)

    def kill(self, *args):
        if self.killed:
            return

        self.killed = True
        self.kill_children(force=True)

    def p(self):
        for t in 'args', 'batches', 'commands':
            print("%s:\n%s\n" % (t, pprint.pformat(getattr(self, t))))


if __name__ == "__main__":
    br = BatchRunner({'data_filename': 'testing'})
    br.read(sys.argv[1])
    br.p()
    br.run_batch('tcpfair')
    pprint.pprint(br.commands_for("tcpfair", 'codel'))



# -*- coding: utf-8 -*-
#
# test_plotters.py
#
# Author:   Toke Høiland-Jørgensen (toke@toke.dk)
# Date:     24 July 2015
# Copyright (c) 2015-2016, Toke Høiland-Jørgensen
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import shutil
import tempfile
import unittest
import traceback

from multiprocessing import Pool
from unittest.util import strclass

from .test_helpers import ForkingTestCase, get_test_data_files

from flent import resultset, formatters
from flent.settings import parser, Settings, DEFAULT_SETTINGS
settings = parser.parse_args(args=[], namespace=Settings(DEFAULT_SETTINGS))

MATPLOTLIB_RC_VALUES = {
    'axes.axisbelow': True,
    'axes.color_cycle': ['#1b9e77', '#d95f02', '#7570b3', '#e7298a',
                         '#66a61e', '#e6ab02', '#a6761d', '#666666'],
    'axes.edgecolor': 'white',
    'axes.facecolor': '#EAEAF2',
    'axes.grid': True,
    'axes.labelcolor': '.15',
    'axes.linewidth': 0,
    'figure.edgecolor': 'white',
    'figure.facecolor': 'white',
    'figure.frameon': False,
    'grid.color': 'white',
    'grid.linestyle': '-',
    'image.cmap': 'Greys',
    'legend.frameon': False,
    'legend.numpoints': 1,
    'legend.scatterpoints': 1,
    'lines.color': '.15',
    'lines.solid_capstyle': 'round',
    'pdf.fonttype': 42,
    'text.color': '.15',
    'xtick.color': '.15',
    'xtick.direction': 'out',
    'xtick.major.size': 0,
    'xtick.minor.size': 0,
    'ytick.color': '.15',
    'ytick.direction': 'out',
    'ytick.major.size': 0,
    'ytick.minor.size': 0,
}

# Some flent test files intentionally lack plots. This list contains the empty
# plots to ensure that they do not fail. The lack of data is either because the
# test did not run with the required flag or because the file is simply older
# than the feature.
MISSING_PLOTS = {
    'test-http-1up.flent.gz': set((
        'tcp_cwnd',
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-http.flent.gz': set((
        'tcp_cwnd',
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-rrul-icmp.flent.gz': set((
        'tcp_cwnd',
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-rrul.flent.gz': set((
        'cpu_core',
        'cpu_core_bar',
        'cpu_core_box',
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-rrul_be-socket_stats.flent.gz': set((
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-rtt-fair.flent.gz': set((
        'tcp_cwnd',
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-tcp_nup.flent.gz': set((
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-tcp_1up_noping-cpu_stats.flent.gz': set((
        'tcp_cwnd',
        'tcp_pacing',
        'tcp_rtt',
        'tcp_rtt_cdf',
        'tcp_rtt_box_combine',
        'tcp_rtt_bar_combine',
    )),
    'test-voip-1up.flent.gz': set((
        'tcp_cwnd',
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
    'test-voip-rrul.flent.gz': set((
        'tcp_cwnd',
        'tcp_delivery_rate',
        'tcp_pacing',
        'tcp_rtt',
        'tcp_rtt_bar_combine',
        'tcp_rtt_box_combine',
        'tcp_rtt_cdf',
    )),
}


class PlottersTestCase(ForkingTestCase):

    def setUp(self):
        from flent import plotters
        if not plotters.HAS_MATPLOTLIB:
            self.skipTest('no matplotlib available')
        self.plotters = plotters

    def __getstate__(self):
        state = {}

        for k, v in self.__dict__.items():
            if k != 'plotters':
                state[k] = v

        return state


class TestPlottersInit(PlottersTestCase):

    def init_test_backend(self, filename):
        self.plotters.init_matplotlib(filename, False, False)
        return self.plotters.matplotlib.get_backend()

    def test_init_fail(self):
        self.plotters.HAS_MATPLOTLIB = False
        self.assertRaises(
            RuntimeError, self.plotters.init_matplotlib, None, None, None)

    def test_init_svg(self):
        self.assertEqual(self.init_test_backend('test.svg'), 'svg')

    def test_init_svgz(self):
        self.assertEqual(self.init_test_backend('test.svgz'), 'svg')

    def test_init_ps(self):
        self.assertEqual(self.init_test_backend('test.ps'), 'ps')

    def test_init_eps(self):
        self.assertEqual(self.init_test_backend('test.eps'), 'ps')

    def test_init_pdf(self):
        self.assertEqual(self.init_test_backend('test.pdf'), 'pdf')

    def test_init_png(self):
        self.assertEqual(self.init_test_backend('test.png'), 'agg')

    def test_init_styles(self):
        self.plotters.init_matplotlib('test.svg', True, False)
        self.assertEqual(len(self.plotters.STYLES), len(self.plotters.LINESTYLES) +
                         len(self.plotters.DASHES) + len(self.plotters.MARKERS))
        for ls in self.plotters.LINESTYLES:
            self.assertIn(dict(linestyle=ls), self.plotters.STYLES)
        for d in self.plotters.DASHES:
            self.assertIn(dict(dashes=d), self.plotters.STYLES)
        for m in self.plotters.MARKERS:
            self.assertIn(dict(marker=m, markevery=10), self.plotters.STYLES)

    def test_init_styles_nomarkers(self):
        self.plotters.init_matplotlib('test.svg', False, False)
        self.assertEqual(len(self.plotters.STYLES), len(self.plotters.LINESTYLES) +
                         len(self.plotters.DASHES))
        for ls in self.plotters.LINESTYLES:
            self.assertIn(dict(linestyle=ls), self.plotters.STYLES)
        for d in self.plotters.DASHES:
            self.assertIn(dict(dashes=d), self.plotters.STYLES)


class TestPlotters(PlottersTestCase):

    def setUp(self):
        super().setUp()
        self.plot_config = {'series': [{'data': 'Test 1'}]}
        self.data_config = {'Test 1': {'units': 'ms'}}
        self.plotters.init_matplotlib('test.svg', True, True)

    def create_plotter(self, plotter_class_name, init=True):
        plotter_class = getattr(self.plotters, plotter_class_name)
        p = plotter_class(self.plot_config, self.data_config)
        self.assertIsInstance(p, plotter_class)
        if init:
            p.init()
        return p

    def test_create_timeseries(self):
        self.create_plotter("TimeseriesPlotter")

    def test_create_timeseries_combine(self):
        self.create_plotter("TimeseriesCombinePlotter")

    def test_create_box(self):
        self.create_plotter("BoxPlotter")

    def test_create_box_combine(self):
        self.create_plotter("BoxCombinePlotter")

    def test_create_bar(self):
        self.create_plotter("BarPlotter")

    def test_create_bar_combine(self):
        self.create_plotter("BarCombinePlotter")

    def test_create_cdf(self):
        self.create_plotter("CdfPlotter")

    def test_create_cdf_combine(self):
        self.create_plotter("CdfCombinePlotter")

    def test_create_qq(self):
        # QQ plots only work with only 1 data series
        p = self.create_plotter("QqPlotter")
        self.plot_config['series'].append({'data': 'Test 1'})

        p = self.create_plotter("QqPlotter", init=False)
        self.assertRaises(RuntimeError, p.init)

    def test_create_ellipsis(self):
        # Ellipsis plots only work with >=2 data series
        p = self.create_plotter("EllipsisPlotter", init=False)
        self.assertRaises(RuntimeError, p.init)

        self.plot_config['series'].append({'data': 'Test 1'})
        p = self.create_plotter("EllipsisPlotter")

    def test_create_subplot_combine(self):
        self.create_plotter("SubplotCombinePlotter")


class TestPlotting(PlottersTestCase):

    def __init__(self, filename, fmt):
        self.filename = filename
        self.fmt = fmt
        unittest.TestCase.__init__(self)

    def setUp(self):
        super().setUp()
        self.output_dir = tempfile.mkdtemp()
        self.settings = settings.copy()

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def __str__(self):
        return "%s - %s (%s)" % (os.path.basename(self.filename), self.fmt,
                                 strclass(self.__class__))

    def runTest(self):
        r = resultset.load(self.filename)
        self.settings.update(r.meta())
        self.settings.load_test(informational=True)
        self.settings.compute_missing_results(r)
        self.settings.FORMAT = 'plot'

        for p in self.settings.PLOTS.keys():
            try:
                self.settings.PLOT = p
                self.settings.OUTPUT = os.path.join(
                    self.output_dir, "%s.%s" % (p, self.fmt))
                formatter = formatters.new(self.settings)
                formatter.format([r])
                res, plen = formatter.verify()
                filename = os.path.basename(self.filename)
                if filename in MISSING_PLOTS and p in MISSING_PLOTS[filename]:
                    continue

                if not res:
                    raise self.failureException(
                        "Verification of plot '%s' failed: %s" % (p, plen))
            except self.failureException:
                raise
            except Exception as e:
                tb = traceback.format_exc()
                new_exc = Exception("Error creating plot '%s'" % p)
                new_exc.orig_tb = tb
                raise new_exc

def plot_one(settings, plot, results):
    from flent import plotters
    plotters.init_matplotlib("-", False, True)
    settings.PLOT = plot
    return plotters.draw_worker(settings, [results])

class TestGUIPlotting(PlottersTestCase):

    def __init__(self, filename):
        self.filename = filename
        unittest.TestCase.__init__(self)

    def setUp(self):
        super().setUp()
        self.output_dir = tempfile.mkdtemp()
        self.settings = settings.copy()

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def __str__(self):
        return "%s - GUI (%s)" % (os.path.basename(self.filename),
                                  strclass(self.__class__))

    def runTest(self):
        results = resultset.load(self.filename)
        self.settings.update(results.meta())
        self.settings.load_test(informational=True)

        with Pool() as pool:
            for p in self.settings.PLOTS.keys():
                plot = pool.apply(plot_one, (self.settings, p, results))
                res, plen = plot.verify()

                filename = os.path.basename(self.filename)
                if filename in MISSING_PLOTS and p in MISSING_PLOTS[filename]:
                    continue

                if not res:
                    raise self.failureException(
                        "Verification of plot '%s' failed: %s" % (p, plen))


dirname = os.path.join(os.path.dirname(__file__), "test_data")
output_formats = ['svg', 'pdf', 'png']
plot_suite = unittest.TestSuite()
for fname in get_test_data_files():
    plot_suite.addTest(TestGUIPlotting(fname))
    for fmt in output_formats:
        plot_suite.addTest(TestPlotting(fname, fmt))


test_suite = unittest.TestSuite(
    [unittest.TestLoader().loadTestsFromTestCase(TestPlottersInit),
     unittest.TestLoader().loadTestsFromTestCase(TestPlotters)])
